from typing import Dict, Type

from spotipyio.testing.components import AlbumsInfoTestComponent
from spotipyio.testing.infra import BaseTestManager, BaseTestComponent


class AlbumsTestManager(BaseTestManager):
    def __init__(self, info: AlbumsInfoTestComponent):
        super().__init__()
        self.info = info

    @staticmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        return {
            "info": AlbumsInfoTestComponent,
        }



from random import choice
from typing import List, Optional, Dict

from pytest_httpserver import RequestHandler, HTTPServer

from spotipyio.logic.consts.spotify_consts import PLAYLISTS, TRACKS, URIS, POSITION, SNAPSHOT_ID
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.logic.internal_models import ChunkSize
from spotipyio.testing.infra import BaseTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory
from spotipyio.logic.internal_tools import DataChunksGenerator


class PlaylistItemsAdderTestComponent(BaseTestComponent):
    def __init__(
        self, server: HTTPServer, headers: Dict[str, str], chunks_generator: DataChunksGenerator = DataChunksGenerator()
    ):
        super().__init__(server=server, headers=headers)
        self._chunks_generator = chunks_generator

    def expect(self, playlist_id: str, uris: List[str], position: Optional[int] = None) -> List[RequestHandler]:
        return self._create_request_handlers(playlist_id=playlist_id, uris=uris, position=position)

    def expect_success(self, playlist_id: str, uris: List[str], position: Optional[int] = None) -> None:
        request_handlers = self._create_request_handlers(playlist_id=playlist_id, uris=uris, position=position)
        self._set_handlers_success(request_handlers)

    def expect_failure(
        self,
        playlist_id: str,
        uris: List[str],
        position: Optional[int] = None,
        response_json: Optional[Json] = None,
        status: Optional[int] = None,
    ) -> None:
        request_handlers = self._create_request_handlers(playlist_id=playlist_id, uris=uris, position=position)
        failed_handler = choice(request_handlers)
        successful_handlers = [handler for handler in request_handlers if handler != failed_handler]
        self._set_handlers_success(successful_handlers)
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)

        failed_handler.respond_with_json(status=status, response_json=response_json)

    def _create_request_handlers(
        self, playlist_id: str, uris: List[str], position: Optional[int]
    ) -> List[RequestHandler]:
        handlers = []

        for chunk in self._chunks_generator.generate_data_chunks(lst=uris, chunk_size=ChunkSize.ITEMS_ADDITION.value):
            payload = {URIS: chunk, POSITION: position}
            handler = self._expect_post_request(route=f"/{PLAYLISTS}/{playlist_id}/{TRACKS}", payload=payload)
            handlers.append(handler)

            if position is not None:
                position += ChunkSize.ITEMS_ADDITION.value

        return handlers

    @staticmethod
    def _set_handlers_success(request_handlers: List[RequestHandler]) -> None:
        for handler in request_handlers:
            response = {SNAPSHOT_ID: SpotifyMockFactory.snapshot_id()}
            handler.respond_with_json(response_json=response, status=201)



from typing import List, Optional

from pytest_httpserver import RequestHandler

from spotipyio.logic.consts.spotify_consts import TIME_RANGE, LIMIT
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.logic.collectors.top_items_collectors.items_type import ItemsType
from spotipyio.logic.collectors.top_items_collectors.time_range import TimeRange
from spotipyio.testing.infra.base_test_component import BaseTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class TopItemsTestComponent(BaseTestComponent):
    def expect(self, items_type: ItemsType, time_range: TimeRange, limit: int = 50) -> List[RequestHandler]:
        request_handler = self._create_request_handler(items_type=items_type, time_range=time_range, limit=limit)
        return [request_handler]

    def expect_success(
        self, items_type: ItemsType, time_range: TimeRange, response_json: Optional[Json] = None, limit: int = 50
    ) -> None:
        handler = self._create_request_handler(items_type=items_type, time_range=time_range, limit=limit)

        if response_json is None:
            response_json = SpotifyMockFactory.user_top_items(items_type)

        handler.respond_with_json(response_json)

    def expect_failure(
        self,
        items_type: ItemsType,
        time_range: TimeRange,
        limit: int = 50,
        status: Optional[int] = None,
        response_json: Optional[int] = None,
    ) -> None:
        handler = self._create_request_handler(items_type=items_type, time_range=time_range, limit=limit)
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)

        handler.respond_with_json(status=status, response_json=response_json)

    def _create_request_handler(self, items_type: ItemsType, time_range: TimeRange, limit: int) -> RequestHandler:
        params = {TIME_RANGE: time_range.value, LIMIT: limit}
        return self._expect_get_request(route=f"/me/top/{items_type.value}", params=params, encode=True)



from spotipyio.logic.collectors.chunks_collectors.albums_collector import AlbumsCollector
from spotipyio.logic.collectors.chunks_collectors.artists_collector import ArtistsCollector
from spotipyio.logic.collectors.chunks_collectors.audio_features_collector import AudioFeaturesCollector
from spotipyio.logic.collectors.chunks_collectors.chapters_collector import ChaptersCollector
from spotipyio.logic.collectors.chunks_collectors.episodes_collector import EpisodesCollector
from spotipyio.logic.collectors.chunks_collectors.tracks_collector import TracksCollector

__all__ = [
    "AlbumsCollector",
    "ArtistsCollector",
    "AudioFeaturesCollector",
    "ChaptersCollector",
    "EpisodesCollector",
    "TracksCollector",
]



from http.server import SimpleHTTPRequestHandler
from typing import Optional


class RedirectURIRequestHandler(SimpleHTTPRequestHandler):
    access_code: Optional[str] = None

    def do_GET(self):
        if self.path.startswith("/?code"):
            RedirectURIRequestHandler.access_code = self.path.split("?code=")[-1]
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(
                b"<html><body><h1>Successfully fetched access code! You may close this window</h1></body></html>"
            )

        else:
            self.send_response(404)
            self.end_headers()



from typing import Dict, Type

from spotipyio.testing.components import SearchItemTestComponent
from spotipyio.testing.infra import BaseTestManager, BaseTestComponent


class SearchTestManager(BaseTestManager):
    def __init__(self, search_item: SearchItemTestComponent):
        super().__init__()
        self.search_item = search_item

    @staticmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        return {
            "search_item": SearchItemTestComponent,
        }



from abc import ABC, abstractmethod
from http import HTTPStatus
from random import choice
from typing import Optional, List, Tuple, Dict
from urllib.parse import urlencode

from multidict import CIMultiDict
from pytest_httpserver import HTTPServer
from pytest_httpserver.httpserver import HandlerType, RequestHandler, UNDEFINED

from spotipyio.logic.consts.typing_consts import Json

INVALID_RESPONSES = {
    HTTPStatus.UNAUTHORIZED: "Unauthorized",
    HTTPStatus.FORBIDDEN: "Bad OAuth Request",
    HTTPStatus.TOO_MANY_REQUESTS: "Too Many Requests",
}


class BaseTestComponent(ABC):
    def __init__(self, server: HTTPServer, headers: CIMultiDict[str]):
        self._server = server
        self._headers = headers
        self._base_url = self._server.url_for("").rstrip("/")

    @abstractmethod
    def expect(self, *args, **kwargs) -> List[RequestHandler]:
        raise NotImplementedError

    @abstractmethod
    def expect_success(self, *args, **kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def expect_failure(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def _expect_get_request(self, route: str, params: Optional[dict] = None, encode: bool = False) -> RequestHandler:
        if params is not None and encode:
            params = urlencode(params)

        return self._server.expect_request(
            uri=route, query_string=params, method="GET", handler_type=HandlerType.ONESHOT, headers=self._headers
        )

    def _expect_post_request(self, route: str, payload: dict) -> RequestHandler:
        return self._server.expect_request(
            uri=route, method="POST", json=payload, handler_type=HandlerType.ONESHOT, headers=self._headers
        )

    def _expect_delete_request(self, route: str, payload: dict) -> RequestHandler:
        return self._server.expect_request(
            uri=route, method="DELETE", json=payload, handler_type=HandlerType.ONESHOT, headers=self._headers
        )

    def _expect_put_request(
        self, route: str, data: Optional[str] = None, payload: Optional[dict] = None
    ) -> RequestHandler:
        return self._server.expect_request(
            uri=route,
            method="PUT",
            data=data,
            json=payload or UNDEFINED,
            handler_type=HandlerType.ONESHOT,
            headers=self._headers,
        )

    def _create_invalid_response(
        self, status: Optional[int] = None, response_json: Optional[Json] = None
    ) -> Tuple[int, Json]:
        if status is None:
            if response_json is None:
                return self._generate_random_status_and_message()
            else:
                status = choice(list(INVALID_RESPONSES.keys()))
                return status.value, response_json

        if response_json is None:
            response_json = choice(list(INVALID_RESPONSES.values()))

        return status, response_json

    @staticmethod
    def _generate_random_status_and_message() -> Tuple[int, Dict[str, dict]]:
        status, message = choice(list(INVALID_RESPONSES.items()))
        json_response = {"error": {"status": status.value, "message": message}}

        return status.value, json_response



from spotipyio.testing.managers.albums_test_manager import AlbumsTestManager
from spotipyio.testing.managers.artists_test_manager import ArtistsTestManager
from spotipyio.testing.managers.chapters_test_manager import ChaptersTestManager
from spotipyio.testing.managers.current_user_test_manager import CurrentUserTestManager
from spotipyio.testing.managers.episodes_test_manager import EpisodesTestManager
from spotipyio.testing.managers.playlists_test_manager import PlaylistsTestManager
from spotipyio.testing.managers.search_test_manager import SearchTestManager
from spotipyio.testing.managers.tracks_test_manager import TracksTestManager
from spotipyio.testing.managers.users_test_manager import UsersTestManager

__all__ = [
    "AlbumsTestManager",
    "ArtistsTestManager",
    "ChaptersTestManager",
    "CurrentUserTestManager",
    "EpisodesTestManager",
    "PlaylistsTestManager",
    "SearchTestManager",
    "TracksTestManager",
    "UsersTestManager",
]



from enum import Enum
from typing import TypeVar, Callable, Any, Awaitable, Union

AF = TypeVar("AF", bound=Callable[..., Awaitable[Any]])
F = TypeVar("F", bound=Callable[..., Any])
Json = Union[dict, list]
EnumType = TypeVar("EnumType", bound=Enum)



from typing import List, Optional

from pytest_httpserver import RequestHandler

from spotipyio.logic.consts.spotify_consts import PLAYLISTS, USERS
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.models import PlaylistCreationRequest
from spotipyio.testing.infra import BaseTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class PlaylistsCreatorTestComponent(BaseTestComponent):
    def expect(self, request: PlaylistCreationRequest) -> List[RequestHandler]:
        return [self._create_request_handler(request)]

    def expect_success(self, request: PlaylistCreationRequest, response_json: Optional[Json] = None) -> None:
        request_handler = self._create_request_handler(request)
        response = response_json or SpotifyMockFactory.playlist(
            user_id=request.user_id, name=request.name, public=request.public, description=request.description
        )

        request_handler.respond_with_json(response_json=response, status=201)

    def expect_failure(
        self, request: PlaylistCreationRequest, status: Optional[int] = None, response_json: Optional[Json] = None
    ) -> None:
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)
        request_handler = self._create_request_handler(request)

        request_handler.respond_with_json(status=status, response_json=response_json)

    def _create_request_handler(self, request: PlaylistCreationRequest) -> RequestHandler:
        return self._expect_post_request(route=f"/{USERS}/{request.user_id}/{PLAYLISTS}", payload=request.to_payload())



import json
from typing import Optional, Dict

from redis import Redis

from spotipyio.auth import ISessionCacheHandler


class RedisSessionCacheHandler(ISessionCacheHandler):
    def __init__(self, key: str, redis: Redis):
        self._key = key
        self._redis = redis

    def get(self) -> Optional[Dict[str, str]]:
        encoded_response = self._redis.get(self._key)

        if encoded_response is not None:
            return json.loads(encoded_response)

    def set(self, response: Dict[str, str]) -> None:
        encoded_response = json.dumps(response)
        self._redis.set(name=self._key, value=encoded_response)



from spotipyio.logic.consts.spotify_consts import PLAYLISTS, USERS
from spotipyio.logic.contract import ISpotifyComponent
from spotipyio.models import PlaylistCreationRequest


class PlaylistsCreator(ISpotifyComponent):
    async def run(self, request: PlaylistCreationRequest) -> dict:
        url = self._url_format.format(user_id=request.user_id)
        return await self._session.post(url=url, payload=request.to_payload())

    @property
    def _url_format(self) -> str:
        return f"{self._base_url}/{USERS}/{{user_id}}/{PLAYLISTS}"



from spotipyio.models.entity_type import EntityType
from spotipyio.models.matching_entity import MatchingEntity
from spotipyio.models.playlist_creation_request import PlaylistCreationRequest
from spotipyio.models.search.search_item import SearchItem
from spotipyio.models.search.search_item_filters import SearchItemFilters
from spotipyio.models.search.search_item_metadata import SearchItemMetadata
from spotipyio.models.search.spotify_search_type import SpotifySearchType

__all__ = [
    "EntityType",
    "MatchingEntity",
    "PlaylistCreationRequest",
    # Search
    "SearchItem",
    "SearchItemFilters",
    "SearchItemMetadata",
    "SpotifySearchType",
]



from spotipyio.logic.updaters.playlist.playlist_items_adder import PlaylistItemsAdder
from spotipyio.logic.updaters.playlist.playlist_items_remover import PlaylistItemsRemover
from spotipyio.logic.updaters.playlist.playlist_items_reorder import PlaylistItemsReorder
from spotipyio.logic.updaters.playlist.playlist_items_replacer import PlaylistItemsReplacer
from spotipyio.logic.updaters.playlist.playlist_cover_updater import PlaylistCoverUpdater

__all__ = [
    "PlaylistCoverUpdater",
    "PlaylistItemsAdder",
    "PlaylistItemsRemover",
    "PlaylistItemsReorder",
    "PlaylistItemsReplacer",
]



import math
import os.path
from io import BytesIO
from sys import getsizeof
from tempfile import TemporaryDirectory
from typing import Optional

from PIL.Image import open as open_image, Image

from spotipyio.logic.consts.image_consts import RGB
from spotipyio.logic.internal_tools.logging import logger
from spotipyio.logic.utils import get_current_timestamp, read_image


class ImageCompressor:
    def compress(
        self, image: bytes, target_size_in_kb: int = 256, quality_interval: int = 5, file_type: str = "jpeg"
    ) -> Optional[bytes]:
        with TemporaryDirectory() as dir_path:
            logger.info(f"Starting to compress image. Target size: {target_size_in_kb}")
            serialized_image = self._serialize_image(image)
            compressed_image_path = self._apply_compression_loop(
                image=serialized_image,
                dir_path=dir_path,
                target_size_in_kb=target_size_in_kb,
                quality_interval=quality_interval,
                file_type=file_type,
            )

            return self._read_compressed_image(compressed_image_path)

    @staticmethod
    def _serialize_image(image: bytes) -> Image:
        image_io = BytesIO(image)
        serialized_image = open_image(image_io)

        if serialized_image.mode.lower() != RGB:
            serialized_image = serialized_image.convert(RGB)

        return serialized_image

    def _apply_compression_loop(
        self, image: Image, dir_path: str, target_size_in_kb: int, quality_interval: int, file_type: str
    ) -> Optional[str]:
        quality = 100
        image_path = None
        image_size = math.inf

        while image_size > target_size_in_kb and quality > 0:
            logger.debug(f"Compressing image using {quality} quality ratio")
            timestamp = get_current_timestamp()
            image_path = os.path.join(dir_path, f"{timestamp}.{file_type}")
            image.save(image_path, quality=quality)
            image_size = self._get_encoded_image_size_in_kb(image_path)
            quality -= quality_interval

        return self._get_compressed_image_path(quality, image_path, image_size)

    @staticmethod
    def _get_encoded_image_size_in_kb(image_path: str) -> float:
        image = read_image(image_path)
        return getsizeof(image) / 1024

    @staticmethod
    def _get_compressed_image_path(quality: int, image_path: str, image_size: float) -> Optional[str]:
        if quality > 0:
            logger.info(
                f"Successfully compressed image using {quality} quality ratio to size of {image_size} KB. Returning "
                f"saved image path"
            )
            return image_path

        else:
            logger.warn("Could not compress image below requested target size. Returning None instead.")

    @staticmethod
    def _read_compressed_image(path: Optional[str]) -> Optional[bytes]:
        if path is None:
            return

        with open(path, "rb") as f:
            return f.read()



from itertools import chain
from random import choice
from typing import Union, List, Iterator, Optional, Any, Type

from spotipyio.logic.consts.typing_consts import EnumType


def chain_iterable(iterable_of_iterable: Union[List[list], Iterator[list]]) -> list:
    return list(chain.from_iterable(iterable_of_iterable))


def safe_nested_get(dct: dict, paths: list, default: Optional[Any] = None) -> Any:
    value = dct.get(paths[0], {})

    for path in paths[1:]:
        value = value.get(path, {})

    return value if value else default


def random_enum_value(enum_: Type[EnumType]) -> EnumType:
    enum_values = get_all_enum_values(enum_)
    return choice(enum_values)


def get_all_enum_values(enum_: Type[EnumType]) -> List[EnumType]:
    return [v for v in enum_]



from spotipyio.testing.components.current_user.current_profile_test_component import CurrentProfileTestComponent
from spotipyio.testing.components.current_user.top_items_test_component import TopItemsTestComponent


__all__ = ["CurrentProfileTestComponent", "TopItemsTestComponent"]



from typing import Dict, Type

from spotipyio.testing.components import TracksInfoTestComponent, TracksAudioFeaturesTestComponent
from spotipyio.testing.infra import BaseTestManager, BaseTestComponent


class TracksTestManager(BaseTestManager):
    def __init__(self, audio_features: TracksAudioFeaturesTestComponent, info: TracksInfoTestComponent):
        super().__init__()
        self.audio_features = audio_features
        self.info = info

    @staticmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        return {
            "audio_features": TracksAudioFeaturesTestComponent,
            "info": TracksInfoTestComponent,
        }



from spotipyio.logic.consts.spotify_consts import ARTISTS
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.logic.internal_models import ChunkSize
from spotipyio.testing.infra.base_chunks_test_component import BaseChunksTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class ArtistsInfoTestComponent(BaseChunksTestComponent):
    @property
    def _route(self) -> str:
        return f"/{ARTISTS}"

    @property
    def _chunk_size(self) -> ChunkSize:
        return ChunkSize.ARTISTS

    @staticmethod
    def _random_valid_response() -> Json:
        return SpotifyMockFactory.several_artists()



from spotipyio.logic.internal_tools.data_chunks_generator import DataChunksGenerator
from spotipyio.logic.internal_tools.image_compressor import ImageCompressor
from spotipyio.logic.internal_tools.logging import logger
from spotipyio.logic.internal_tools.pool_executor import PoolExecutor

__all__ = ["DataChunksGenerator", "ImageCompressor", "logger", "PoolExecutor"]



from random import choice
from typing import List, Optional

from multidict import CIMultiDict
from pytest_httpserver import RequestHandler, HTTPServer

from spotipyio.logic.consts.spotify_consts import PLAYLISTS, TRACKS, SNAPSHOT_ID, URI
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.logic.internal_models import ChunkSize
from spotipyio.logic.internal_tools import DataChunksGenerator
from spotipyio.testing.infra import BaseTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class PlaylistsItemsRemoverTestComponent(BaseTestComponent):
    def __init__(
        self, server: HTTPServer, headers: CIMultiDict, chunks_generator: DataChunksGenerator = DataChunksGenerator()
    ):
        super().__init__(server=server, headers=headers)
        self._chunks_generator = chunks_generator

    def expect(self, playlist_id: str, uris: List[str], snapshot_id: str) -> List[RequestHandler]:
        chunks = self._to_chunks(uris)
        snapshots_ids = self._create_snapshots_ids(
            chunks=chunks,
            provided_snapshot=snapshot_id,
        )
        return self._create_request_handlers(
            playlist_id=playlist_id,
            chunks=chunks,
            snapshots_ids=snapshots_ids,
        )

    def expect_success(
        self, playlist_id: str, uris: List[str], snapshot_id: str, expected_snapshots: Optional[List[str]] = None
    ) -> None:
        chunks = self._to_chunks(uris)
        snapshots_ids = self._create_snapshots_ids(
            chunks=chunks, provided_snapshot=snapshot_id, expected_snapshots=expected_snapshots
        )
        request_handlers = self._create_request_handlers(
            playlist_id=playlist_id,
            chunks=chunks,
            snapshots_ids=snapshots_ids,
        )

        self._set_handlers_success(request_handlers, snapshots_ids)

    def expect_failure(
        self,
        playlist_id: str,
        uris: List[str],
        snapshot_id: str,
        response_json: Optional[Json] = None,
        status: Optional[int] = None,
    ) -> None:
        chunks = self._to_chunks(uris)
        snapshots_ids = self._create_snapshots_ids(
            chunks=chunks,
            provided_snapshot=snapshot_id,
        )
        request_handlers = self._create_request_handlers(
            playlist_id=playlist_id,
            chunks=chunks,
            snapshots_ids=snapshots_ids,
        )
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)

        self._set_handlers_failure(
            request_handlers=request_handlers, snapshots_ids=snapshots_ids, response_json=response_json, status=status
        )

    def _create_request_handlers(
        self, playlist_id: str, chunks: List[List[str]], snapshots_ids: List[str]
    ) -> List[RequestHandler]:
        handlers = []

        for (
            i,
            chunk,
        ) in enumerate(chunks):
            payload = {TRACKS: [{URI: uri} for uri in chunk], SNAPSHOT_ID: snapshots_ids[i]}
            handler = self._expect_delete_request(route=f"/{PLAYLISTS}/{playlist_id}/{TRACKS}", payload=payload)
            handlers.append(handler)

        return handlers

    def _create_snapshots_ids(
        self, chunks: List[List[str]], provided_snapshot: str, expected_snapshots: Optional[List[str]] = None
    ) -> List[str]:
        if expected_snapshots:
            self._validate_expected_snapshots_number(chunks, expected_snapshots)
            snapshots_ids = expected_snapshots
        else:
            snapshots_ids = [SpotifyMockFactory.snapshot_id() for _ in chunks]

        return [provided_snapshot] + snapshots_ids

    @staticmethod
    def _validate_expected_snapshots_number(chunks: List[List[str]], expected_snapshots: List[str]) -> None:
        snapshots_number = len(expected_snapshots)
        requests_number = len(chunks)

        if snapshots_number != requests_number:
            raise ValueError(
                f"Received {snapshots_number} expected snapshots while expecting only {requests_number} requests"
            )

    @staticmethod
    def _set_handlers_success(request_handlers: List[RequestHandler], snapshots_ids: List[str]) -> None:
        for i, handler in enumerate(request_handlers):
            snapshot_id = snapshots_ids[i + 1]
            response = {SNAPSHOT_ID: snapshot_id}
            handler.respond_with_json(response_json=response, status=200)

    def _set_handlers_failure(
        self, request_handlers: List[RequestHandler], snapshots_ids: List[str], response_json: Json, status: int
    ) -> None:
        failed_handler_index = choice(list(range(len(request_handlers))))

        for i, handler in enumerate(request_handlers):
            if i == failed_handler_index:
                handler.respond_with_json(response_json=response_json, status=status)
            else:
                self._set_handlers_success(request_handlers=[handler], snapshots_ids=snapshots_ids[i : i + 2])

    def _to_chunks(self, uris: List[str]) -> List[List[str]]:
        chunks = self._chunks_generator.generate_data_chunks(lst=uris, chunk_size=ChunkSize.ITEMS_REMOVAL.value)
        return list(chunks)



from typing import Dict, Type

from spotipyio.testing.components import ChaptersInfoTestComponent
from spotipyio.testing.infra import BaseTestManager, BaseTestComponent


class ChaptersTestManager(BaseTestManager):
    def __init__(self, info: ChaptersInfoTestComponent):
        super().__init__()
        self.info = info

    @staticmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        return {
            "info": ChaptersInfoTestComponent,
        }



from spotipyio.logic.consts.spotify_consts import TIME_RANGE, LIMIT
from spotipyio.logic.contract import ISpotifyComponent
from spotipyio.logic.collectors.top_items_collectors.items_type import ItemsType
from spotipyio.logic.collectors.top_items_collectors.time_range import TimeRange


class TopItemsCollector(ISpotifyComponent):
    async def run(self, items_type: ItemsType, time_range: TimeRange, limit: int = 50) -> dict:
        url = self._url_format.format(items_type=items_type.value)
        params = {TIME_RANGE: time_range.value, LIMIT: limit}
        return await self._session.get(url=url, params=params)

    @property
    def _url_format(self) -> str:
        return f"{self._base_url}/me/top/{{items_type}}"



from abc import ABC, abstractmethod
from typing import List

from spotipyio.logic.consts.spotify_consts import IDS
from spotipyio.logic.contract.spotify_component_interface import ISpotifyComponent
from spotipyio.auth import SpotifySession
from spotipyio.logic.internal_models import ChunkSize
from spotipyio.logic.internal_tools import DataChunksGenerator
from spotipyio.logic.utils.general_utils import chain_iterable


class BaseChunksCollector(ISpotifyComponent, ABC):
    def __init__(
        self, base_url: str, session: SpotifySession, chunks_generator: DataChunksGenerator = DataChunksGenerator()
    ):
        super().__init__(base_url=base_url, session=session)
        self._chunks_generator = chunks_generator
        self._formatted_route = self._route.replace("-", "_")

    async def run(self, ids: List[str]) -> List[dict]:
        chunks = await self._chunks_generator.execute_by_chunk_in_parallel(
            lst=ids, func=self._run_single, expected_type=list, chunk_size=self._chunk_size.value
        )
        return chain_iterable(chunks)

    async def _run_single(self, ids: List[str]) -> List[dict]:
        response = await self._session.get(url=self._url, params={IDS: ",".join(ids)})
        return response[self._formatted_route]

    @property
    @abstractmethod
    def _chunk_size(self) -> ChunkSize:
        raise NotImplementedError

    @property
    @abstractmethod
    def _route(self) -> str:
        raise NotImplementedError

    @property
    def _url(self) -> str:
        return f"{self._base_url}/{self._route}"



from random import randint
from typing import List, Dict

from spotipyio.logic.consts.spotify_consts import (
    HREF,
    LIMIT,
    NEXT,
    OFFSET,
    PREVIOUS,
    TOTAL,
    ITEMS,
    ALBUMS,
    ARTISTS,
    AUDIOBOOKS,
    EPISODES,
    PLAYLISTS,
    SHOWS,
    TRACKS,
)
from spotipyio.models import SpotifySearchType, SearchItem


class SearchResponseBuilder:
    def __init__(self, search_item: SearchItem):
        self._search_item = search_item
        self._search_response = {}

    def add(self, search_type: SpotifySearchType, items: List[dict]) -> None:
        response = self._build_single_search_response(search_type)
        response[ITEMS] = items
        key = self._search_type_key_map[search_type]

        self._search_response[key] = response

    def build(self) -> dict:
        return self._search_response

    def _build_single_search_response(self, search_type: SpotifySearchType) -> dict:
        return {
            HREF: self._build_search_response_href(search_type, offset=0),
            LIMIT: 20,
            NEXT: self._build_search_response_href(search_type, offset=20),
            OFFSET: 0,
            PREVIOUS: None,
            TOTAL: randint(1, 1000),
        }

    def _build_search_response_href(self, search_type: SpotifySearchType, offset: int) -> str:
        query = self._search_item.to_query_params([search_type])
        return f"https://api.spotify.com/v1/search?query={query}&offset={offset}&limit=20"

    @property
    def _search_type_key_map(self) -> Dict[SpotifySearchType, str]:
        return {
            SpotifySearchType.ALBUM: ALBUMS,
            SpotifySearchType.ARTIST: ARTISTS,
            SpotifySearchType.AUDIOBOOK: AUDIOBOOKS,
            SpotifySearchType.EPISODE: EPISODES,
            SpotifySearchType.PLAYLIST: PLAYLISTS,
            SpotifySearchType.SHOW: SHOWS,
            SpotifySearchType.TRACK: TRACKS,
        }



from typing import List, Optional

from multidict import CIMultiDict
from pytest_httpserver import RequestHandler, HTTPServer

from spotipyio.logic.consts.spotify_consts import PLAYLISTS, TRACKS, OFFSET, LIMIT, ADDITIONAL_TYPES, TRACK
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.testing.infra import BaseTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory
from spotipyio.testing.utils.random_paged_responses_builder import RandomPagedResponsesBuilder


class PlaylistsInfoTestComponent(BaseTestComponent):
    def __init__(self, server: HTTPServer, headers: CIMultiDict):
        super().__init__(server=server, headers=headers)
        self._paged_responses_builder = RandomPagedResponsesBuilder(base_url=self._base_url, page_max_size=100)

    def expect(self, id_: str, expected_pages: int = 1) -> List[RequestHandler]:
        return self._create_request_handlers(id_, expected_pages)

    def expect_success(self, id_: str, response_jsons: Optional[List[Json]] = None, expected_pages: int = 1) -> None:
        request_handlers = self._create_request_handlers(id_, expected_pages)
        responses = self._build_responses(
            provided_responses=response_jsons,
            request_handlers=request_handlers,
            playlist_id=id_,
            expected_pages=expected_pages,
        )

        for handler, response in zip(request_handlers, responses):
            handler.respond_with_json(response)

    def expect_failure(
        self, id_: str, status: Optional[int] = None, response_json: Optional[Json] = None, expected_pages: int = 1
    ) -> None:
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)
        request_handlers = self._create_request_handlers(id_, expected_pages)
        first_handler = request_handlers[0]

        first_handler.respond_with_json(status=status, response_json=response_json)

    def _create_request_handlers(self, id_: str, expected_pages: int) -> List[RequestHandler]:
        handlers = [self._expect_get_request(route=f"/{PLAYLISTS}/{id_}")]

        for i in range(1, expected_pages):
            params = {OFFSET: str(i * 100), LIMIT: "100", ADDITIONAL_TYPES: TRACK}
            page_handler = self._expect_get_request(route=f"/{PLAYLISTS}/{id_}/{TRACKS}", params=params)
            handlers.append(page_handler)

        return handlers

    def _build_responses(
        self,
        provided_responses: Optional[List[Json]],
        request_handlers: List[RequestHandler],
        playlist_id: str,
        expected_pages: int,
    ) -> List[Json]:
        if provided_responses is not None:
            self._validate_provided_responses(provided_responses, request_handlers)
            return provided_responses

        return self._paged_responses_builder.build(playlist_id, expected_pages)

    @staticmethod
    def _validate_provided_responses(provided_responses: List[Json], request_handlers: List[RequestHandler]) -> None:
        request_handlers_number = len(request_handlers)
        provided_responses_number = len(provided_responses)

        if request_handlers_number != provided_responses_number:
            raise ValueError(f"Expected {request_handlers_number} responses but got {provided_responses_number}")

    def _generate_random_responses(self, request_handlers: List[RequestHandler], playlist_id: str) -> List[Json]:
        handlers_number = len(request_handlers)

        if handlers_number == 1:
            next_url = None
        else:
            next_url = f"{self._base_url}/{PLAYLISTS}/{playlist_id}/{TRACKS}?{OFFSET}=100&{LIMIT}=100"

        tracks = SpotifyMockFactory.playlist_tracks(next=next_url)
        responses = [SpotifyMockFactory.playlist(id=playlist_id, tracks=tracks)]

        for _ in range(1, handlers_number):
            page_response = SpotifyMockFactory.playlist_tracks(entity_id=playlist_id)
            responses.append(page_response)

        return responses



from spotipyio.testing.components.albums import AlbumsInfoTestComponent
from spotipyio.testing.components.artists import ArtistsInfoTestComponent, ArtistsTopTracksTestComponent
from spotipyio.testing.components.chapters import ChaptersInfoTestComponent
from spotipyio.testing.components.current_user import CurrentProfileTestComponent, TopItemsTestComponent
from spotipyio.testing.components.episodes import EpisodesInfoTestComponent
from spotipyio.testing.components.playlists import (
    PlaylistsCoverUpdaterTestComponent,
    PlaylistsCreatorTestComponent,
    PlaylistsInfoTestComponent,
    PlaylistItemsAdderTestComponent,
    PlaylistsItemsRemoverTestComponent,
)
from spotipyio.testing.components.search import SearchItemTestComponent
from spotipyio.testing.components.tracks import TracksAudioFeaturesTestComponent, TracksInfoTestComponent
from spotipyio.testing.components.users import UserPlaylistsTestComponent


__all__ = [
    # Albums
    "AlbumsInfoTestComponent",
    # Artists
    "ArtistsInfoTestComponent",
    "ArtistsTopTracksTestComponent",
    # Chapter
    "ChaptersInfoTestComponent",
    # Current User
    "CurrentProfileTestComponent",
    "TopItemsTestComponent",
    # Episodes
    "EpisodesInfoTestComponent",
    # Playlists
    "PlaylistsCoverUpdaterTestComponent",
    "PlaylistsCreatorTestComponent",
    "PlaylistsInfoTestComponent",
    "PlaylistItemsAdderTestComponent",
    "PlaylistsItemsRemoverTestComponent",
    # Search
    "SearchItemTestComponent",
    # Tracks
    "TracksAudioFeaturesTestComponent",
    "TracksInfoTestComponent",
    # Users
    "UserPlaylistsTestComponent",
]



from spotipyio.tools.extractors.entity_extractor_interface import IEntityExtractor
from spotipyio.tools.extractors.primary_artist_entity_extractor import PrimaryArtistEntityExtractor
from spotipyio.tools.extractors.track_entity_extractor import TrackEntityExtractor

__all__ = [
    "IEntityExtractor",
    "PrimaryArtistEntityExtractor",
    "TrackEntityExtractor",
]



from typing import Type, Dict

from spotipyio.logic.contract import BaseManager, ISpotifyComponent
from spotipyio.logic.collectors import ArtistsCollector, ArtistsTopTracksCollector


class ArtistsManager(BaseManager):
    def __init__(self, info: ArtistsCollector, top_tracks: ArtistsTopTracksCollector):
        super().__init__()
        self.info = info
        self.top_tracks = top_tracks

    @staticmethod
    def _components() -> Dict[str, Type[ISpotifyComponent]]:
        return {"info": ArtistsCollector, "top_tracks": ArtistsTopTracksCollector}



from dataclasses import dataclass
from typing import Optional


@dataclass
class SearchItemFilters:
    track: Optional[str] = None
    artist: Optional[str] = None
    album: Optional[str] = None
    year: Optional[int] = None
    # TODO: Add missing fields options



from base64 import b64encode


def read_image(path: str) -> bytes:
    with open(path, "rb") as f:
        return f.read()


def encode_image_to_base64(image: bytes) -> str:
    encoded_image = b64encode(image)
    return encoded_image.decode("utf-8")



from dataclasses import dataclass, fields, field
from typing import Optional, Dict, List
from urllib.parse import quote

from spotipyio.models.search.spotify_search_type import SpotifySearchType
from spotipyio.models.search.search_item_filters import SearchItemFilters
from spotipyio.models.search.search_item_metadata import SearchItemMetadata


@dataclass
class SearchItem:
    text: Optional[str] = None
    filters: SearchItemFilters = field(default_factory=lambda: SearchItemFilters())
    metadata: SearchItemMetadata = field(default_factory=lambda: SearchItemMetadata())

    def __post_init__(self):
        self._validate_input()

    def to_query_params(self, search_types: Optional[List[SpotifySearchType]] = None) -> Dict[str, str]:
        if search_types:
            types = [search_type.value for search_type in search_types]
        else:
            types = [search_type.value for search_type in self.metadata.search_types]

        return {"q": self._build_query(), "type": ",".join(types)}

    def _build_query(self) -> str:
        query_components = self._get_query_component()
        query = " ".join(query_components)

        if self.metadata.quote:
            return quote(query)

        return query

    def _get_query_component(self) -> List[str]:
        query_components = []

        if self.text is not None:
            query_components.append(self.text)

        for filter_field in fields(self.filters):
            field_value = getattr(self.filters, filter_field.name)

            if field_value is not None:
                query_components.append(f"{filter_field.name}:{field_value}")

        return query_components

    def _validate_input(self) -> None:
        filters_values = [getattr(self.filters, field.name) for field in fields(self.filters)]
        are_all_filters_missing = all(value is None for value in filters_values)

        if self.text is None and are_all_filters_missing:
            raise ValueError("You must supply text or at least one search filter")



from typing import Dict, Type

from spotipyio.testing.components import ArtistsInfoTestComponent, ArtistsTopTracksTestComponent
from spotipyio.testing.infra import BaseTestManager, BaseTestComponent


class ArtistsTestManager(BaseTestManager):
    def __init__(self, info: ArtistsInfoTestComponent, top_tracks: ArtistsTopTracksTestComponent):
        super().__init__()
        self.info = info
        self.top_tracks = top_tracks

    @staticmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        return {"info": ArtistsInfoTestComponent, "top_tracks": ArtistsTopTracksTestComponent}



from abc import ABC, abstractmethod
from typing import List, Optional, Dict

from pytest_httpserver import HTTPServer, RequestHandler

from spotipyio.logic.consts.spotify_consts import IDS
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.logic.internal_models import ChunkSize
from spotipyio.testing.infra import BaseTestComponent
from spotipyio.logic.internal_tools import DataChunksGenerator


class BaseChunksTestComponent(BaseTestComponent, ABC):
    def __init__(
        self, server: HTTPServer, headers: Dict[str, str], chunks_generator: DataChunksGenerator = DataChunksGenerator()
    ):
        super().__init__(server=server, headers=headers)
        self._chunks_generator = chunks_generator

    def expect(self, ids: List[str]) -> List[RequestHandler]:
        return self._expect_chunks(route=self._route, ids=ids, chunk_size=self._chunk_size.value)

    def expect_failure(
        self, ids: List[str], status: Optional[int] = None, response_json: Optional[Json] = None
    ) -> None:
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)
        handlers = self._expect_chunks(route=self._route, ids=ids, chunk_size=self._chunk_size.value)

        for handler in handlers:
            handler.respond_with_json(status=status, response_json=response_json)

    def expect_success(self, ids: List[str], responses_json: Optional[List[Json]] = None) -> None:
        handlers = self._expect_chunks(route=self._route, ids=ids, chunk_size=self._chunk_size.value)
        handlers_number = len(handlers)
        responses = responses_json or [self._random_valid_response() for _ in range(handlers_number)]
        responses_number = len(responses)

        if responses_number != handlers_number:
            raise ValueError(
                f"Number of provided responses ({responses_number}) didn't match number of handlers ({handlers_number})"
            )

        for handler, response_json in zip(handlers, responses):
            handler.respond_with_json(status=200, response_json=response_json)

    def _expect_chunks(self, route: str, ids: List[str], chunk_size: int) -> List[RequestHandler]:
        chunks = self._chunks_generator.generate_data_chunks(lst=ids, chunk_size=chunk_size)
        request_handlers = []

        for chunk in chunks:
            chunk_handler = self._expect_get_request(route=route, params={IDS: ",".join(chunk)})
            request_handlers.append(chunk_handler)

        return request_handlers

    @property
    @abstractmethod
    def _chunk_size(self) -> ChunkSize:
        raise NotImplementedError

    @property
    @abstractmethod
    def _route(self) -> str:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _random_valid_response() -> Json:
        raise NotImplementedError



from typing import Type, Dict

from spotipyio.logic.contract import BaseManager, ISpotifyComponent
from spotipyio.logic.collectors import UserPlaylistsCollector


class UsersManager(BaseManager):
    def __init__(self, playlists: UserPlaylistsCollector):
        super().__init__()
        self.playlists = playlists

    @staticmethod
    def _components() -> Dict[str, Type[ISpotifyComponent]]:
        return {
            "playlists": UserPlaylistsCollector,
        }



from spotipyio.logic.contract import ISpotifyComponent


class CurrentProfileCollector(ISpotifyComponent):
    async def run(self):
        return await self._session.get(url=self._url)

    @property
    def _url(self) -> str:
        return f"{self._base_url}/me"



from abc import ABC, abstractmethod
from functools import partial
from typing import List, Optional

from spotipyio.logic.contract.spotify_component_interface import ISpotifyComponent
from spotipyio.auth import SpotifySession
from spotipyio.logic.internal_tools import PoolExecutor


class BasePaginationCollector(ISpotifyComponent, ABC):
    def __init__(self, base_url: str, session: SpotifySession, pool_executor: PoolExecutor = PoolExecutor()):
        super().__init__(base_url=base_url, session=session)
        self._pool_executor = pool_executor

    async def run(self, ids: List[str], max_pages: int = 1) -> List[dict]:
        func = partial(self.run_single, max_pages=max_pages)
        return await self._pool_executor.run(iterable=ids, func=func, expected_type=dict)

    async def run_single(self, id_: str, max_pages: int = 1) -> dict:
        url = self._url_format.format(id=id_)
        result = await self._session.get(url=url)

        if max_pages > 1:
            await self._append_additional_pages_items(result, max_pages)

        return result

    async def _append_additional_pages_items(self, result: dict, max_pages: int) -> None:
        current_page = 2
        next_url = self._extract_first_next_url(result)

        while next_url is not None and current_page <= max_pages:
            page = await self._session.get(url=next_url, params=self._additional_items_request_params)
            self._extend_existing_items(result, page)
            next_url = self._extract_subsequent_next_url(page)
            current_page += 1

    @property
    @abstractmethod
    def _url_format(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def _additional_items_request_params(self) -> Optional[dict]:
        raise NotImplementedError

    @abstractmethod
    def _extract_first_next_url(self, result: dict) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def _extract_subsequent_next_url(self, page: dict) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def _extend_existing_items(self, result: dict, page: dict) -> None:
        raise NotImplementedError



from typing import List, Optional, Dict

from spotipyio.logic.consts.spotify_consts import PLAYLISTS, TRACKS, NEXT, ITEMS, TRACK, ADDITIONAL_TYPES
from spotipyio.logic.contract import BasePaginationCollector
from spotipyio.logic.utils import safe_nested_get


class PlaylistsCollector(BasePaginationCollector):
    @property
    def _url_format(self) -> str:
        return f"{self._base_url}/{PLAYLISTS}/{{id}}"

    @property
    def _additional_items_request_params(self) -> Dict[str, str]:
        return {ADDITIONAL_TYPES: TRACK}

    def _extract_first_next_url(self, result: dict) -> Optional[str]:
        return safe_nested_get(result, [TRACKS, NEXT])

    def _extract_subsequent_next_url(self, page: dict) -> Optional[str]:
        return page[NEXT]

    def _extend_existing_items(self, result: dict, page: dict) -> None:
        existing_items: List[dict] = safe_nested_get(result, [TRACKS, ITEMS])

        if existing_items:
            page_items = page[ITEMS]
            existing_items.extend(page_items)



from typing import Optional

from spotipyio.logic.consts.spotify_consts import NAME
from spotipyio.tools.extractors.entity_extractor_interface import IEntityExtractor


class TrackEntityExtractor(IEntityExtractor):
    def extract(self, entity: dict) -> Optional[str]:
        return entity.get(NAME)

    @property
    def name(self) -> str:
        return "track"



from spotipyio.logic.consts.spotify_consts import AUDIO_FEATURES_ROUTE
from spotipyio.logic.contract import BaseChunksCollector
from spotipyio.logic.internal_models import ChunkSize


class AudioFeaturesCollector(BaseChunksCollector):
    @property
    def _route(self) -> str:
        return AUDIO_FEATURES_ROUTE

    @property
    def _chunk_size(self) -> ChunkSize:
        return ChunkSize.AUDIO_FEATURES



from spotipyio.testing.components.playlists.playlist_cover_updater_test_component import (
    PlaylistsCoverUpdaterTestComponent,
)
from spotipyio.testing.components.playlists.playlists_creator_test_component import PlaylistsCreatorTestComponent
from spotipyio.testing.components.playlists.playlists_info_test_component import PlaylistsInfoTestComponent
from spotipyio.testing.components.playlists.playlists_items_adder_test_component import PlaylistItemsAdderTestComponent
from spotipyio.testing.components.playlists.playlists_items_remover_test_component import (
    PlaylistsItemsRemoverTestComponent,
)

__all__ = [
    "PlaylistsCoverUpdaterTestComponent",
    "PlaylistsCreatorTestComponent",
    "PlaylistsInfoTestComponent",
    "PlaylistItemsAdderTestComponent",
    "PlaylistsItemsRemoverTestComponent",
]



from dataclasses import fields, Field
from typing import Tuple, List, Optional, Dict, Any

from spotipyio.tools.extractors import IEntityExtractor, TrackEntityExtractor, PrimaryArtistEntityExtractor
from spotipyio.models import MatchingEntity
from spotipyio.logic.utils import compute_similarity_score


class EntityMatcher:
    def __init__(
        self,
        extractors: Optional[Dict[IEntityExtractor, float]] = None,
        threshold: float = 0.7,
        min_present_fields: int = 2,
    ):
        self._validate_extractors_scores(extractors)
        self._extractors = extractors or self._get_default_extractors()
        self._threshold = threshold
        self._min_present_fields = min_present_fields

    def match(self, entity: MatchingEntity, candidate: Any) -> Tuple[bool, float]:
        scores = []

        for field in fields(entity):
            field_score = self._compute_field_score(entity, field, candidate)

            if field_score is not None:
                scores.append(field_score)

        return self._is_matching(scores)

    def _compute_field_score(self, entity: MatchingEntity, field: Field, candidate: Any) -> Optional[float]:
        extractor = self._names_to_extractors.get(field.name)
        if extractor is None:
            return

        extractor_score = self._names_to_scores[field.name]
        entity = getattr(entity, field.name)

        if entity is not None:
            score = self._compute_raw_field_score(extractor, candidate, entity)

            if score is not None:
                return score * extractor_score

    @staticmethod
    def _compute_raw_field_score(extractor: IEntityExtractor, raw_candidate: Any, entity: str) -> Optional[float]:
        if raw_candidate is not None:
            candidate = extractor.extract(raw_candidate)

            if candidate is not None:
                return compute_similarity_score(candidate, entity)

    @staticmethod
    def _get_default_extractors() -> Dict[IEntityExtractor, float]:
        return {TrackEntityExtractor(): 0.65, PrimaryArtistEntityExtractor(): 0.35}

    def _is_matching(self, scores: List[float]) -> Tuple[bool, float]:
        if len(scores) < self._min_present_fields:
            return False, -1

        weighted_score = sum(scores)
        is_matching = weighted_score >= self._threshold

        return is_matching, weighted_score

    @staticmethod
    def _validate_extractors_scores(extractors: Optional[Dict[IEntityExtractor, float]]) -> None:
        if extractors is not None:
            extractors_total_score = sum(extractors.values())

            if extractors_total_score != 1:
                raise ValueError(
                    f"The sum of scores of given extractors is {extractors_total_score}. entity extractors' scores "
                    f"must sum up to exactly 1"
                )

    @property
    def _names_to_extractors(self) -> Dict[str, IEntityExtractor]:
        return {extractor.name: extractor for extractor in self._extractors.keys()}

    @property
    def _names_to_scores(self) -> Dict[str, float]:
        return {extractor.name: score for extractor, score in self._extractors.items()}



from dataclasses import dataclass, field
from typing import List

from spotipyio.logic.utils import get_all_enum_values
from spotipyio.models.search.spotify_search_type import SpotifySearchType


@dataclass
class SearchItemMetadata:
    search_types: List[SpotifySearchType] = field(default_factory=lambda: get_all_enum_values(SpotifySearchType))
    quote: bool = True

    def __post_init__(self):
        if not self.search_types:
            raise ValueError("SearchItemMetadata must include at least one SpotifySearchType")



from spotipyio.logic.contract.base_manager import BaseManager
from spotipyio.logic.contract.collectors.base_chunks_collector import BaseChunksCollector
from spotipyio.logic.contract.collectors.base_pagination_collector import BasePaginationCollector

__all__ = [
    "BaseChunksCollector",
    "BaseManager",
    "BasePaginationCollector",
    "BasePlaylistsUpdater",
    "BaseSingleCollector",
    "ISpotifyComponent",
]

from spotipyio.logic.contract.collectors.base_single_collector import BaseSingleCollector
from spotipyio.logic.contract.spotify_component_interface import ISpotifyComponent

from spotipyio.logic.contract.updaters.base_playlist_updater import BasePlaylistsUpdater



from dataclasses import dataclass, fields
from typing import Optional


@dataclass
class MatchingEntity:
    track: Optional[str] = None
    artist: Optional[str] = None

    def __post_init__(self):
        self._validate_input()

    def _validate_input(self) -> None:
        if all(getattr(self, field.name) is None for field in fields(self)):
            raise ValueError("At least one matching candidate field must be supplied")



from spotipyio.testing.components.tracks.tracks_audio_features_test_component import TracksAudioFeaturesTestComponent
from spotipyio.testing.components.tracks.tracks_info_test_component import TracksInfoTestComponent

__all__ = [
    "TracksAudioFeaturesTestComponent",
    "TracksInfoTestComponent",
]



from typing import Type, Dict

from spotipyio.logic.contract import BaseManager, ISpotifyComponent
from spotipyio.logic.collectors import CurrentProfileCollector, TopItemsCollector


class CurrentUserManager(BaseManager):
    def __init__(self, top_items: TopItemsCollector, profile: CurrentProfileCollector):
        super().__init__()
        self.top_items = top_items
        self.profile = profile

    @staticmethod
    def _components() -> Dict[str, Type[ISpotifyComponent]]:
        return {"top_items": TopItemsCollector, "profile": CurrentProfileCollector}



from typing import Type, Dict

from spotipyio.testing.components import (
    PlaylistsCreatorTestComponent,
    PlaylistsInfoTestComponent,
    PlaylistItemsAdderTestComponent,
    PlaylistsItemsRemoverTestComponent,
    PlaylistsCoverUpdaterTestComponent,
)
from spotipyio.testing.infra import BaseTestManager, BaseTestComponent


class PlaylistsTestManager(BaseTestManager):
    def __init__(
        self,
        add_items: PlaylistItemsAdderTestComponent,
        create: PlaylistsCreatorTestComponent,
        info: PlaylistsInfoTestComponent,
        remove_items: PlaylistsItemsRemoverTestComponent,
        update_cover: PlaylistsCoverUpdaterTestComponent,
    ):
        super().__init__()
        self.add_items = add_items
        self.create = create
        self.info = info
        self.remove_items = remove_items
        self.update_cover = update_cover

    @staticmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        return {
            "add_items": PlaylistItemsAdderTestComponent,
            "create": PlaylistsCreatorTestComponent,
            "info": PlaylistsInfoTestComponent,
            "remove_items": PlaylistsItemsRemoverTestComponent,
            "update_cover": PlaylistsCoverUpdaterTestComponent,
        }



from spotipyio.testing.components.artists.artists_info_test_component import ArtistsInfoTestComponent
from spotipyio.testing.components.artists.artists_top_tracks_test_component import ArtistsTopTracksTestComponent

__all__ = ["ArtistsInfoTestComponent", "ArtistsTopTracksTestComponent"]



from abc import ABC, abstractmethod
from typing import Type, Dict

from spotipyio.logic.contract.spotify_component_interface import ISpotifyComponent
from spotipyio.auth import SpotifySession


class BaseManager(ABC):
    def __init__(self, **named_component: Dict[str, ISpotifyComponent]):
        pass

    @classmethod
    def create(cls, base_url: str, session: SpotifySession) -> "BaseManager":
        named_components = {}

        for name, component in cls._components().items():
            named_components[name] = component(base_url=base_url, session=session)

        return cls(**named_components)

    @staticmethod
    @abstractmethod
    def _components() -> Dict[str, Type[ISpotifyComponent]]:
        raise NotImplementedError



from typing import List

from spotipyio.logic.consts.spotify_consts import TRACKS, URIS
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.logic.contract import BasePlaylistsUpdater


class PlaylistItemsReplacer(BasePlaylistsUpdater):
    async def run(self, playlist_id: str, uris: List[str]) -> Json:
        url = self._build_url(playlist_id)
        return await self._session.put(url=url, payload={URIS: uris})

    @property
    def _route(self) -> str:
        return TRACKS



from abc import ABC, abstractmethod
from typing import Any, Optional


class IEntityExtractor(ABC):
    @abstractmethod
    def extract(self, entity: Any) -> Optional[str]:
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError



from __future__ import annotations
from typing import Optional, Dict, Any

from aiohttp import ClientSession, ClientResponse, ContentTypeError, ClientResponseError
from multidict import CIMultiDict

from spotipyio.auth.client_credentials import ClientCredentials
from spotipyio.logic.consts.api_consts import ACCESS_TOKEN, REFRESH_TOKEN, TOKEN_REQUEST_URL
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.auth.session_cache_handler_interface import ISessionCacheHandler
from spotipyio.logic.authorization import AccessTokenGenerator
from spotipyio.auth.spotify_grant_type import SpotifyGrantType
from spotipyio.logic.utils.web_utils import create_client_session


class SpotifySession:
    def __init__(
        self,
        token_request_url: str = TOKEN_REQUEST_URL,
        credentials: Optional[ClientCredentials] = None,
        access_token_generator: Optional[AccessTokenGenerator] = None,
        session: Optional[ClientSession] = None,
        session_cache_handler: Optional[ISessionCacheHandler] = None,
    ):
        self._token_request_url = token_request_url
        self._credentials = credentials or ClientCredentials()
        self._access_token_generator = access_token_generator
        self._session = session
        self._cache_handler = session_cache_handler

    async def get(self, url: str, params: Optional[dict] = None) -> Optional[Json]:
        async with self._session.get(url=url, params=params) as response:
            return await self._handle_response(response)

    async def post(self, url: str, payload: dict) -> Optional[Json]:
        async with self._session.post(url=url, json=payload) as response:
            return await self._handle_response(response)

    async def put(self, url: str, data: Optional[Any] = None, payload: Optional[dict] = None) -> Optional[Json]:
        async with self._session.put(url=url, data=data, json=payload) as response:
            return await self._handle_response(response)

    async def delete(self, url: str, payload: Optional[dict] = None) -> Optional[Json]:
        async with self._session.delete(url=url, json=payload) as response:
            return await self._handle_response(response)

    async def refresh(self) -> None:
        if self._session is not None:
            await self._session.close()

        self._session = await self._build_client_session(use_cache=False)

    async def start(self) -> None:
        if self._session is None:
            self._session = await self._build_client_session(use_cache=True)

    async def stop(self) -> None:
        if self._session is not None:
            await self._session.close()

        if self._access_token_generator is not None:
            await self._access_token_generator.stop()

    def get_authorization_headers(self) -> CIMultiDict[str]:
        return self._session.headers

    async def _handle_response(self, response: ClientResponse) -> Optional[Json]:
        if self._is_2xx_successful(response):
            return await self._jsonify_response_if_possible(response)

        json_error_response = await self._jsonify_response_if_possible(response)
        if json_error_response is None:
            response.raise_for_status()

        raise ClientResponseError(
            request_info=response.request_info,
            history=response.history,
            status=response.status,
            message=f"Spotify request to URL `{response.request_info.url}` with method "
            f"`{response.request_info.method}` failed with the following JSON message:\n{json_error_response}",
        )

    @staticmethod
    def _is_2xx_successful(response: ClientResponse) -> bool:
        return 200 <= response.status < 300

    @staticmethod
    async def _jsonify_response_if_possible(response: ClientResponse) -> Optional[Json]:
        try:
            return await response.json()

        except ContentTypeError:
            return

    async def __aenter__(self) -> SpotifySession:
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.stop()

    async def _build_client_session(self, use_cache: bool) -> ClientSession:
        if self._access_token_generator is None:
            await self._init_token_generator()

        response = await self._fetch_access_token(use_cache)
        access_token = response[ACCESS_TOKEN]
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
        }

        return create_client_session(headers)

    async def _fetch_access_token(self, use_cache: bool) -> Dict[str, str]:
        if use_cache and self._cache_handler is not None:
            response = await self._retrieve_access_token_from_cache()

            if response is not None:
                return response

        return await self._generate_access_token(
            grant_type=self._credentials.grant_type, access_code=self._credentials.access_code
        )

    async def _retrieve_access_token_from_cache(self) -> Optional[Dict[str, str]]:
        cached_response = self._cache_handler.get()

        if cached_response is not None:
            return await self._generate_access_token(
                grant_type=SpotifyGrantType.REFRESH_TOKEN, access_code=cached_response[REFRESH_TOKEN]
            )

    async def _generate_access_token(self, grant_type: SpotifyGrantType, access_code: str) -> Dict[str, str]:
        response = await self._access_token_generator.generate(grant_type=grant_type, access_code=access_code)

        if self._cache_handler is not None:
            self._cache_handler.set(response)

        return response

    async def _init_token_generator(self) -> None:
        self._access_token_generator = AccessTokenGenerator(
            token_request_url=self._token_request_url,
            client_id=self._credentials.client_id,
            client_secret=self._credentials.client_secret,
            redirect_uri=self._credentials.redirect_uri,
        )
        await self._access_token_generator.start()



from typing import List

from spotipyio.logic.consts.spotify_consts import TRACKS, URI, SNAPSHOT_ID
from spotipyio.logic.contract import BasePlaylistsUpdater
from spotipyio.auth import SpotifySession
from spotipyio.logic.internal_models import ChunkSize
from spotipyio.logic.internal_tools import DataChunksGenerator


class PlaylistItemsRemover(BasePlaylistsUpdater):
    def __init__(
        self, base_url: str, session: SpotifySession, chunks_generator: DataChunksGenerator = DataChunksGenerator()
    ):
        super().__init__(base_url=base_url, session=session)
        self._chunks_generator = chunks_generator

    async def run(self, playlist_id: str, uris: List[str], snapshot_id: str) -> str:
        chunks = self._chunks_generator.generate_data_chunks(lst=uris, chunk_size=ChunkSize.ITEMS_REMOVAL.value)
        url = self._build_url(playlist_id)

        for chunk in chunks:
            snapshot_id = await self._remove_single_chunk(url=url, uris=chunk, snapshot_id=snapshot_id)

        return snapshot_id

    async def _remove_single_chunk(self, url: str, snapshot_id: str, uris: List[str]) -> str:
        payload = {TRACKS: [{URI: uri} for uri in uris], SNAPSHOT_ID: snapshot_id}
        response = await self._session.delete(url=url, payload=payload)

        return response[SNAPSHOT_ID]

    @property
    def _route(self) -> str:
        return TRACKS



from abc import ABC, abstractmethod

from spotipyio.logic.consts.spotify_consts import PLAYLISTS
from spotipyio.logic.contract import ISpotifyComponent


class BasePlaylistsUpdater(ISpotifyComponent, ABC):
    @property
    @abstractmethod
    def _route(self) -> str:
        raise NotImplementedError

    def _build_url(self, playlist_id: str) -> str:
        return f"{self._base_url}/{PLAYLISTS}/{playlist_id}/{self._route}"



SPOTIFY_API_BASE_URL = "https://api.spotify.com/v1"
ID = "id"
IDS = "ids"
TRACK = "track"
TRACKS = "tracks"
ITEMS = "items"
TIME_RANGE = "time_range"
LIMIT = "limit"
NAME = "name"
DESCRIPTION = "description"
PUBLIC = "public"
URI = "uri"
URIS = "uris"
PLAYLISTS = "playlists"
IMAGES = "images"
ARTISTS = "artists"
POSITION = "position"
RANGE_START = "range_start"
INSERT_BEFORE = "insert_before"
SNAPSHOT_ID = "snapshot_id"
RANGE_LENGTH = "range_length"
NEXT = "next"
ADDITIONAL_TYPES = "additional_types"
USERS = "users"
OWNER = "owner"
HREF = "href"
OFFSET = "offset"
TOTAL = "total"
ALBUMS = "albums"
AUDIO_FEATURES_ROUTE = "audio-features"
AUDIO_FEATURES = "audio_features"
CHAPTERS = "chapters"
EPISODES = "episodes"
SHOWS = "shows"
AUDIOBOOKS = "audiobooks"
PREVIOUS = "previous"



from typing import Optional

from spotipyio.logic.consts.spotify_consts import NEXT, ITEMS, PLAYLISTS, USERS
from spotipyio.logic.contract import BasePaginationCollector


class UserPlaylistsCollector(BasePaginationCollector):
    @property
    def _url_format(self) -> str:
        return f"{self._base_url}/{USERS}/{{id}}/{PLAYLISTS}?offset=0&limit=50"

    @property
    def _additional_items_request_params(self) -> None:
        return None

    def _extract_first_next_url(self, result: dict) -> Optional[str]:
        return result[NEXT]

    def _extract_subsequent_next_url(self, page: dict) -> Optional[str]:
        return page[NEXT]

    def _extend_existing_items(self, result: dict, page: dict) -> None:
        existing_items = result[ITEMS]
        page_items = page[ITEMS]
        existing_items.extend(page_items)



from abc import ABC, abstractmethod
from typing import Dict, Optional


class ISessionCacheHandler(ABC):
    @abstractmethod
    def get(self) -> Optional[Dict[str, str]]:
        raise NotImplementedError

    @abstractmethod
    def set(self, response: Dict[str, str]) -> None:
        raise NotImplementedError



from __future__ import annotations
from typing import Dict, Optional

from aiohttp import ClientSession

from spotipyio.auth.spotify_grant_type import SpotifyGrantType
from spotipyio.logic.authorization.authorization_payload_builder import AuthorizationPayloadBuilder
from spotipyio.logic.utils import create_client_session, encode_bearer_token


class AccessTokenGenerator:
    def __init__(
        self,
        token_request_url: str,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        redirect_uri: Optional[str] = None,
        session: Optional[ClientSession] = None,
    ):
        self._token_request_url = token_request_url
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._session = session

    async def generate(self, grant_type: SpotifyGrantType, access_code: Optional[str]) -> Dict[str, str]:
        encoded_header = encode_bearer_token(client_id=self._client_id, client_secret=self._client_secret)
        headers = {"Authorization": f"Basic {encoded_header}"}
        data = AuthorizationPayloadBuilder.build(
            grant_type=grant_type, access_code=access_code, client_id=self._client_id, redirect_uri=self._redirect_uri
        )

        async with self._session.post(url=self._token_request_url, headers=headers, data=data) as raw_response:
            raw_response.raise_for_status()
            return await raw_response.json()

    async def start(self) -> None:
        if self._session is None:
            raw_session = create_client_session()
            self._session = await raw_session.__aenter__()

    async def stop(self) -> None:
        if self._session is not None:
            await self._session.close()

    async def __aenter__(self) -> AccessTokenGenerator:
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.stop()



from typing import Type, Dict

from spotipyio.logic.collectors import SearchCollector
from spotipyio.logic.contract import BaseManager, ISpotifyComponent


class SearchManager(BaseManager):
    def __init__(self, search_item: SearchCollector):
        super().__init__()
        self.search_item = search_item

    @staticmethod
    def _components() -> Dict[str, Type[ISpotifyComponent]]:
        return {
            "search_item": SearchCollector,
        }



TOKEN_REQUEST_URL = "https://accounts.spotify.com/api/token"
ACCESS_CODE_REQUEST_URL = "https://accounts.spotify.com/authorize/"
REDIRECT_URI = "redirect_uri"
CODE = "code"
GRANT_TYPE = "grant_type"
JSON = "json"
REFRESH_TOKEN = "refresh_token"
CLIENT_ID = "client_id"
ACCESS_TOKEN = "access_token"
EXPIRES_IN = "expires_in"
SCOPE = "scope"
TOKEN_TYPE = "token_type"
RESPONSE_TYPE = "response_type"



from functools import partial
from typing import Sized, Any, Awaitable, Callable, List, Optional, Type

from asyncio_pool import AioPool
from tqdm import tqdm

from spotipyio.logic.internal_tools.logging import logger


class PoolExecutor:
    def __init__(self, pool_size: int = 5, validate_results: bool = True):
        self._pool_size = pool_size
        self._validate_results = validate_results

    async def run(
        self, iterable: Sized, func: Callable[..., Awaitable[Any]], expected_type: Optional[Type] = None
    ) -> List[Any]:
        if not iterable:
            logger.warning("PoolExecutor did not receive any values in iterable. Returning empty list by default")
            return []

        return await self._execute_in_pool(iterable=iterable, func=func, expected_type=expected_type)

    async def _execute_in_pool(
        self, iterable: Sized, func: Callable[..., Awaitable[Any]], expected_type: Optional[Type] = None
    ) -> List[Any]:
        pool = AioPool(self._pool_size)

        with tqdm(total=len(iterable)) as progress_bar:
            monitored_func = partial(self._execute_single, progress_bar, func)
            results = await pool.map(monitored_func, iterable)

        if self._validate_results:
            return self._filter_out_invalid_results(results, expected_type)

        return results

    @staticmethod
    async def _execute_single(progress_bar: tqdm, func: Callable[..., Awaitable[Any]], value: Any) -> Any:
        try:
            return await func(value)

        except Exception as e:
            logger.exception("PoolExecutor encountered exception")
            return e

        finally:
            progress_bar.update(1)

    @staticmethod
    def _filter_out_invalid_results(results: List[Any], expected_type: Type) -> List[Any]:
        valid_results = [result for result in results if isinstance(result, expected_type)]
        logger.info(f"Successfully retrieved {len(valid_results)} valid results out of {len(results)} total requests")

        return valid_results



from spotipyio.logic.consts.spotify_consts import TRACKS
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.logic.internal_models import ChunkSize
from spotipyio.testing.infra.base_chunks_test_component import BaseChunksTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class TracksInfoTestComponent(BaseChunksTestComponent):
    @property
    def _route(self) -> str:
        return f"/{TRACKS}"

    @property
    def _chunk_size(self) -> ChunkSize:
        return ChunkSize.TRACKS

    @staticmethod
    def _random_valid_response() -> Json:
        return SpotifyMockFactory.several_tracks()



from calendar import monthrange
from datetime import datetime
from random import randint, choice, random, uniform
from typing import Optional, List, Dict, Callable, Any, Type

from spotipyio.logic.consts.spotify_consts import (
    PLAYLISTS,
    USERS,
    LIMIT,
    HREF,
    NEXT,
    OFFSET,
    TOTAL,
    ITEMS,
    ARTISTS,
    TRACKS,
    ALBUMS,
    TRACK,
    AUDIO_FEATURES,
    CHAPTERS,
    EPISODES,
    SHOWS,
    AUDIOBOOKS,
)
from spotipyio.logic.consts.typing_consts import EnumType
from spotipyio.logic.collectors.top_items_collectors.items_type import ItemsType
from spotipyio.models import SearchItem, SearchItemFilters, SearchItemMetadata, SpotifySearchType
from spotipyio.testing.utils.search_response_builder import SearchResponseBuilder
from spotipyio.logic.utils import random_alphanumeric_string


class SpotifyMockFactory:
    @staticmethod
    def spotify_id() -> str:
        return random_alphanumeric_string(min_length=22, max_length=22)

    @staticmethod
    def some_spotify_ids(length: Optional[int] = None) -> List[str]:
        number_of_ids = length or randint(2, 10)
        return [SpotifyMockFactory.spotify_id() for _ in range(number_of_ids)]

    @staticmethod
    def paged_playlists(**kwargs) -> dict:
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())
        href = kwargs.get("href") or SpotifyMockFactory.href(
            entity_type=USERS, entity_id=entity_id, extra_routes=[PLAYLISTS]
        )
        return {
            HREF: href,
            LIMIT: kwargs.get(LIMIT, randint(1, 50)),
            NEXT: "",
            OFFSET: kwargs.get(OFFSET, randint(1, 200)),
            TOTAL: kwargs.get(TOTAL, randint(1, 1000)),
            ITEMS: kwargs.get(ITEMS, SpotifyMockFactory._some_playlists()),
        }

    @staticmethod
    def playlist(user_id: Optional[str] = None, **kwargs) -> dict:
        owner = kwargs.get("owner", SpotifyMockFactory.owner(user_id))
        entity_type = "playlist"
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())

        return {
            "collaborative": kwargs.get("collaborative", SpotifyMockFactory._random_boolean()),
            "description": kwargs.get("description", random_alphanumeric_string()),
            "external_urls": SpotifyMockFactory.external_urls(entity_type=entity_type, entity_id=entity_id),
            "followers": kwargs.get("followers", SpotifyMockFactory.followers()),
            "href": SpotifyMockFactory.href(entity_type=entity_type, entity_id=entity_id),
            "id": entity_id,
            "images": kwargs.get("images", SpotifyMockFactory.images()),
            "name": kwargs.get("name", SpotifyMockFactory.name()),
            "owner": owner,
            "public": kwargs.get("public", SpotifyMockFactory._random_boolean()),
            "snapshot_id": kwargs.get("snapshot_id", SpotifyMockFactory.snapshot_id()),
            "tracks": kwargs.get("tracks", SpotifyMockFactory.playlist_tracks(id=entity_id, owner=owner)),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type=entity_type, entity_id=entity_id),
            "primary_color": None,
        }

    @staticmethod
    def playlist_tracks(**kwargs) -> dict:
        entity_type = "playlist"
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())
        owner = kwargs.get("owner")
        total_tracks = kwargs.get("total", randint(1, 100))
        items = [SpotifyMockFactory.playlist_item(owner) for _ in range(total_tracks)]

        return {
            "href": SpotifyMockFactory.href(entity_type=entity_type, entity_id=entity_id, extra_routes=["tracks"]),
            "limit": 100,
            "next": kwargs.get("next"),
            "offset": 0,
            "previous": None,
            "total": total_tracks,
            "items": items,
        }

    @staticmethod
    def playlist_item(owner: Optional[str] = None) -> dict:
        added_at = SpotifyMockFactory._random_datetime()
        return {
            "added_at": added_at.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "added_by": owner or SpotifyMockFactory.owner(),
            "is_local": SpotifyMockFactory._random_boolean(),
            "track": SpotifyMockFactory.track(),
        }

    @staticmethod
    def snapshot_id() -> str:
        return random_alphanumeric_string(min_length=32, max_length=32)

    @staticmethod
    def owner(entity_id: Optional[str] = None) -> dict:
        entity_type = "user"
        if entity_id is None:
            entity_id = random_alphanumeric_string()

        return {
            "external_urls": SpotifyMockFactory.external_urls(entity_type=entity_type, entity_id=entity_id),
            "href": SpotifyMockFactory.href(entity_type=entity_type, entity_id=entity_id),
            "id": entity_id,
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type=entity_type, entity_id=entity_id),
            "display": random_alphanumeric_string(),
        }

    @staticmethod
    def several_artists(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.artist, ids=ids, key=ARTISTS)

    @staticmethod
    def several_tracks(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.track, ids=ids, key=TRACKS)

    @staticmethod
    def several_albums(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.album, ids=ids, key=ALBUMS)

    @staticmethod
    def several_chapters(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.chapter, ids=ids, key=CHAPTERS)

    @staticmethod
    def several_episodes(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.episode, ids=ids, key=EPISODES)

    @staticmethod
    def several_audio_features(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.audio_features, ids=ids, key=AUDIO_FEATURES)

    @staticmethod
    def several_shows(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.show, ids=ids, key=SHOWS)

    @staticmethod
    def several_playlists(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.playlist, ids=ids, key=PLAYLISTS)

    @staticmethod
    def several_audiobooks(ids: Optional[List[str]] = None) -> Dict[str, List[dict]]:
        return SpotifyMockFactory._several_items(method=SpotifyMockFactory.audiobook, ids=ids, key=AUDIOBOOKS)

    @staticmethod
    def artist(**kwargs) -> dict:
        entity_type = "artist"
        entity_id = kwargs.get("id") or SpotifyMockFactory.spotify_id()

        return {
            "external_urls": kwargs.get("external_urls") or SpotifyMockFactory.external_urls(entity_type, entity_id),
            "followers": kwargs.get("followers", SpotifyMockFactory.followers()),
            "genres": kwargs.get("genres", SpotifyMockFactory.genres()),
            "href": SpotifyMockFactory.href(entity_type, entity_id),
            "id": entity_id,
            "images": kwargs.get("images", SpotifyMockFactory.images()),
            "name": kwargs.get("name", SpotifyMockFactory.name()),
            "popularity": kwargs.get("popularity", SpotifyMockFactory.popularity()),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type=entity_type, entity_id=entity_id),
        }

    @staticmethod
    def track(**kwargs) -> dict:
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())
        entity_type = "track"
        artists = kwargs.get("artists", SpotifyMockFactory._some_artists())

        return {
            "album": kwargs.get("album", SpotifyMockFactory.album(artists=artists)),
            "artists": artists,
            "available_markets": kwargs.get("available_markets", SpotifyMockFactory.available_markets()),
            "disc_number": kwargs.get("disc_number", SpotifyMockFactory.disc_number()),
            "duration_ms": kwargs.get("duration_ms", SpotifyMockFactory.duration_ms()),
            "explicit": kwargs.get("explicit", SpotifyMockFactory._random_boolean()),
            "external_ids": kwargs.get("external_ids", SpotifyMockFactory.external_ids()),
            "external_urls": kwargs.get("external_urls") or SpotifyMockFactory.external_urls(entity_type, entity_id),
            "href": SpotifyMockFactory.href(entity_type=entity_type, entity_id=entity_id),
            "id": entity_id,
            "is_local": kwargs.get("is_local", SpotifyMockFactory._random_boolean()),
            "is_playable": kwargs.get("is_playable", SpotifyMockFactory._random_boolean()),
            "name": kwargs.get("name", SpotifyMockFactory.name()),
            "popularity": kwargs.get("popularity", SpotifyMockFactory.popularity()),
            "preview_url": kwargs.get("preview_url", SpotifyMockFactory.preview_url()),
            "track_number": kwargs.get("track_number", SpotifyMockFactory.track_number()),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type=entity_type, entity_id=entity_id),
        }

    @staticmethod
    def album(**kwargs) -> dict:
        entity_type = "album"
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())

        return {
            "artists": kwargs.get("artists", SpotifyMockFactory._some_artists()),
            "album_type": kwargs.get("album_type", SpotifyMockFactory.album_type()),
            "total_tracks": kwargs.get("total_tracks", SpotifyMockFactory.track_number()),
            "external_urls": SpotifyMockFactory.external_urls(entity_type=entity_type, entity_id=entity_id),
            "available_markets": kwargs.get("available_markets", SpotifyMockFactory.available_markets()),
            "href": SpotifyMockFactory.href(entity_type=entity_type, entity_id=entity_id),
            "id": entity_id,
            "images": kwargs.get("images", SpotifyMockFactory.images()),
            "name": kwargs.get("name", SpotifyMockFactory.name()),
            "release_date": kwargs.get("release_date", SpotifyMockFactory.release_date()),
            "release_date_precision": kwargs.get("release_date_precision", "day"),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type=entity_type, entity_id=entity_id),
            "is_playable": kwargs.get("is_playable", SpotifyMockFactory._random_boolean()),
        }

    @staticmethod
    def audio_features(**kwargs) -> dict:
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())
        entity_type = TRACK

        return {
            "acousticness": SpotifyMockFactory._random_confidence("acousticness", **kwargs),
            "analysis_url": f"https://api.spotify.com/v1/audio-analysis/{entity_id}",
            "danceability": SpotifyMockFactory._random_confidence("danceability", **kwargs),
            "duration_ms": kwargs.get("duration_ms", randint(90000, 360000)),
            "energy": SpotifyMockFactory._random_confidence("energy", **kwargs),
            "id": entity_id,
            "instrumentalness": SpotifyMockFactory._random_confidence("instrumentalness", **kwargs),
            "key": kwargs.get("key", randint(0, 11)),
            "liveness": SpotifyMockFactory._random_confidence("liveness", **kwargs),
            "loudness": kwargs.get("loudness", randint(-60, 3)),
            "mode": kwargs.get("mode", SpotifyMockFactory._random_boolean()),
            "speechiness": SpotifyMockFactory._random_confidence("speechiness", **kwargs),
            "tempo": kwargs.get("tempo", uniform(40, 200)),
            "time_signature": kwargs.get("time_signature", randint(0, 5)),
            "track_href": SpotifyMockFactory.href(entity_type=entity_type, entity_id=entity_id),
            "type": "audio_features",
            "uri": SpotifyMockFactory.uri(entity_type=entity_type, entity_id=entity_id),
            "valence": SpotifyMockFactory._random_confidence("valence", **kwargs),
        }

    @staticmethod
    def chapter(**kwargs) -> dict:
        entity_type = "episode"
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())
        description = kwargs.get("description", SpotifyMockFactory.description())

        return {
            "audio_preview_url": kwargs.get("audio_preview_url", SpotifyMockFactory.preview_url()),
            "available_markets": kwargs.get("available_markets", SpotifyMockFactory.available_markets()),
            "chapter_number": kwargs.get("chapter_number", randint(1, 20)),
            "description": description,
            "html_description": SpotifyMockFactory.html_description(description),
            "duration_ms": kwargs.get("duration_ms", SpotifyMockFactory.duration_ms()),
            "explicit": kwargs.get("explicit", SpotifyMockFactory._random_boolean()),
            "external_urls": SpotifyMockFactory.external_urls(entity_type, entity_id),
            "href": SpotifyMockFactory.href(entity_type, entity_id),
            "id": entity_id,
            "images": kwargs.get("images", SpotifyMockFactory.images()),
            "is_playable": kwargs.get("is_playable", SpotifyMockFactory._random_boolean()),
            "languages": kwargs.get("languages", SpotifyMockFactory._random_string_array()),
            "name": kwargs.get("name", SpotifyMockFactory.name()),
            "release_date": kwargs.get("release_date", SpotifyMockFactory.release_date()),
            "release_date_precision": "day",
            "resume_point": kwargs.get("resume_point", SpotifyMockFactory.resume_point()),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type, entity_id),
            "restrictions": kwargs.get("restrictions", SpotifyMockFactory.restrictions()),
            "audiobook": kwargs.get("audiobook", SpotifyMockFactory.audiobook()),
        }

    @staticmethod
    def episode(**kwargs) -> dict:
        entity_type = "show"
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())
        description = kwargs.get("description", SpotifyMockFactory.description())

        return {
            "audio_preview_url": kwargs.get("audio_preview_url", SpotifyMockFactory.preview_url()),
            "description": description,
            "html_description": SpotifyMockFactory.html_description(description),
            "duration_ms": kwargs.get("duration_ms", SpotifyMockFactory.duration_ms()),
            "explicit": kwargs.get("explicit", SpotifyMockFactory._random_boolean()),
            "external_urls": SpotifyMockFactory.external_urls(entity_type, entity_id),
            "href": SpotifyMockFactory.href(entity_type, entity_id),
            "id": entity_id,
            "images": kwargs.get("images", SpotifyMockFactory.images()),
            "is_externally_hosted": kwargs.get("is_externally_hosted", SpotifyMockFactory._random_boolean()),
            "is_playable": kwargs.get("is_playable", SpotifyMockFactory._random_boolean()),
            "language": kwargs.get("language", random_alphanumeric_string()),
            "languages": kwargs.get("languages", SpotifyMockFactory._random_string_array()),
            "name": kwargs.get("name", SpotifyMockFactory.name()),
            "release_date": kwargs.get("release_date", SpotifyMockFactory.release_date()),
            "release_date_precision": "day",
            "resume_point": kwargs.get("resume_point", SpotifyMockFactory.resume_point()),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type, entity_id),
            "restrictions": kwargs.get("restrictions", SpotifyMockFactory.restrictions()),
            "show": kwargs.get("show", SpotifyMockFactory.show()),
        }

    @staticmethod
    def description() -> str:
        return random_alphanumeric_string()

    @staticmethod
    def html_description(description: Optional[str] = None) -> str:
        description_text = description or random_alphanumeric_string()
        return f"<p>{description_text}</p>"

    @staticmethod
    def album_type() -> str:
        return choice(["album", "single", "compilation"])

    @staticmethod
    def available_markets() -> List[str]:
        return SpotifyMockFactory._random_string_array()

    @staticmethod
    def disc_number() -> int:
        return randint(1, 2)

    @staticmethod
    def duration_ms() -> int:
        return randint(90000, 360000)

    @staticmethod
    def external_urls(entity_type: str, entity_id: str) -> Dict[str, str]:
        return {"spotify": f"https://open.spotify.com/{entity_type}/{entity_id}"}

    @staticmethod
    def external_ids() -> Dict[str, str]:
        return {"isrc": random_alphanumeric_string()}

    @staticmethod
    def href(entity_type: str, entity_id: str, extra_routes: Optional[List[str]] = None) -> str:
        href = f"https://api.spotify.com/v1/{entity_type}/{entity_id}"

        if extra_routes:
            formatted_routes = [route.strip("/") for route in extra_routes]
            joined_route = "/".join(formatted_routes)
            href += f"/{joined_route}"

        return href

    @staticmethod
    def followers(number: Optional[int] = None) -> dict:
        followers_number = number or randint(500, 50000000)
        return {"href": None, "total": followers_number}

    @staticmethod
    def genres(length: Optional[int] = None) -> List[str]:
        genres_number = length or randint(0, 5)
        return SpotifyMockFactory._random_string_array(genres_number)

    @staticmethod
    def images() -> List[Dict[str, str]]:
        return [SpotifyMockFactory._random_image(size) for size in [640, 320, 160]]

    @staticmethod
    def name() -> str:
        return random_alphanumeric_string()

    @staticmethod
    def popularity() -> int:
        return randint(0, 100)

    @staticmethod
    def preview_url() -> str:
        preview_id = random_alphanumeric_string(min_length=40, max_length=40)
        return f"https://p.scdn.co/mp3-preview/{preview_id}"

    @staticmethod
    def release_date() -> str:
        raw_date = SpotifyMockFactory._random_datetime()
        return raw_date.strftime("%Y-%m-%d")

    @staticmethod
    def track_number() -> int:
        return randint(1, 20)

    @staticmethod
    def uri(entity_type: str, entity_id: str) -> str:
        return f"spotify:{entity_type}:{entity_id}"

    @staticmethod
    def some_uris(entity_type: str, length: Optional[int] = None) -> List[str]:
        number_of_uris = length or randint(2, 10)
        uris = []

        for _ in range(number_of_uris):
            uri = SpotifyMockFactory.uri(entity_type=entity_type, entity_id=SpotifyMockFactory.spotify_id())
            uris.append(uri)

        return uris

    @staticmethod
    def user_profile(entity_id: Optional[str] = None) -> dict:
        entity_type = "user"
        if entity_id is None:
            entity_id = SpotifyMockFactory.spotify_id()

        return {
            "country": random_alphanumeric_string(),
            "display_name": random_alphanumeric_string(),
            "email": random_alphanumeric_string(),
            "explicit_content": {
                "filter_enabled": SpotifyMockFactory._random_boolean(),
                "filter_locked": SpotifyMockFactory._random_boolean(),
            },
            "external_urls": SpotifyMockFactory.external_urls(entity_type=entity_type, entity_id=entity_id),
            "followers": SpotifyMockFactory.followers(),
            "href": SpotifyMockFactory.href(entity_type=entity_type, entity_id=entity_id),
            "id": entity_id,
            "images": SpotifyMockFactory.images(),
            "product": choice(["premium", "free"]),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type=entity_type, entity_id=entity_id),
        }

    @staticmethod
    def user_top_items(items_type: ItemsType) -> dict:
        total_tracks = randint(1, 10)
        href = SpotifyMockFactory.href(entity_type="me", entity_id="", extra_routes=["top", items_type.value])

        if items_type == ItemsType.ARTISTS:
            items = SpotifyMockFactory._some_artists()
        else:
            items = SpotifyMockFactory._some_tracks()

        return {
            "href": href,
            "limit": 50,
            "next": None,
            "offset": 0,
            "previous": None,
            "total": total_tracks,
            "items": items,
        }

    @staticmethod
    def search_item() -> SearchItem:
        search_types = SpotifyMockFactory._random_multi_enum_values(SpotifySearchType) or [SpotifySearchType.TRACK]
        return SearchItem(
            text=random_alphanumeric_string(),
            filters=SearchItemFilters(
                track=SpotifyMockFactory._optional_random_alphanumeric_string(),
                artist=SpotifyMockFactory._optional_random_alphanumeric_string(),
                album=SpotifyMockFactory._optional_random_alphanumeric_string(),
                year=SpotifyMockFactory._an_optional(lambda: SpotifyMockFactory._random_datetime().year),
            ),
            metadata=SearchItemMetadata(
                search_types=search_types,
                quote=SpotifyMockFactory._random_boolean(),
            ),
        )

    @staticmethod
    def search_response(search_item: SearchItem) -> Dict[str, dict]:
        search_types_method_mapping = {
            SpotifySearchType.ALBUM: SpotifyMockFactory.album,
            SpotifySearchType.ARTIST: SpotifyMockFactory.album,
            SpotifySearchType.AUDIOBOOK: SpotifyMockFactory.audiobook,
            SpotifySearchType.EPISODE: SpotifyMockFactory.episode,
            SpotifySearchType.PLAYLIST: SpotifyMockFactory.playlist,
            SpotifySearchType.SHOW: SpotifyMockFactory.show,
            SpotifySearchType.TRACK: SpotifyMockFactory.track,
        }
        builder = SearchResponseBuilder(search_item)

        for search_type in search_item.metadata.search_types:
            search_type_method = search_types_method_mapping[search_type]
            items = SpotifyMockFactory._some_items(search_type_method)
            builder.add(search_type, items)

        return builder.build()

    @staticmethod
    def _several_items(method: Callable[..., dict], ids: Optional[List[str]], key: str) -> Dict[str, List[dict]]:
        if ids:
            items = [method(id=item_id) for item_id in ids]
        else:
            items = SpotifyMockFactory._some_items(method)

        return {key: items}

    @staticmethod
    def _some_items(method: Callable[..., dict]) -> List[dict]:
        return [method() for _ in range(randint(1, 10))]

    @staticmethod
    def _random_image(size: int) -> dict:
        image_id = random_alphanumeric_string(min_length=40, max_length=40)
        return {"url": f"https://i.scdn.co/image/{image_id}", "height": size, "width": size}

    @staticmethod
    def _some_artists() -> List[dict]:
        return SpotifyMockFactory._some_items(SpotifyMockFactory.artist)

    @staticmethod
    def _some_tracks() -> List[dict]:
        return SpotifyMockFactory._some_items(SpotifyMockFactory.track)

    @staticmethod
    def _some_playlists() -> List[dict]:
        return SpotifyMockFactory._some_items(SpotifyMockFactory.playlist)

    @staticmethod
    def _random_string_array(length: Optional[int] = None) -> List[str]:
        n_elements = length or randint(0, 10)
        return [random_alphanumeric_string() for _ in range(n_elements)]

    @staticmethod
    def _random_boolean() -> bool:
        return choice([True, False])

    @staticmethod
    def _random_datetime() -> datetime:
        current_year = datetime.now().year
        year = randint(1950, current_year)
        month = randint(1, 12)
        _, last_month_day = monthrange(year, month)
        day = randint(1, last_month_day)

        return datetime(year, month, day)

    @staticmethod
    def _random_confidence(key: str, **kwargs) -> float:
        return kwargs.get(key, random())

    @staticmethod
    def resume_point(**kwargs) -> dict:
        return {
            "fully_played": kwargs.get("fully_played", SpotifyMockFactory._random_boolean()),
            "resume_position_ms": kwargs.get("resume_position_ms", SpotifyMockFactory.duration_ms()),
        }

    @staticmethod
    def restrictions(**kwargs) -> dict:
        return {"reason": kwargs.get("reason", random_alphanumeric_string())}

    @staticmethod
    def audiobook(**kwargs) -> dict:
        entity_type = "audiobook"
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())
        description = kwargs.get("description", SpotifyMockFactory.description())

        return {
            "authors": kwargs.get("authors", SpotifyMockFactory.some_authors()),
            "available_markets": kwargs.get("available_markets", SpotifyMockFactory.available_markets()),
            "copyrights": kwargs.get("copyrights", SpotifyMockFactory.some_copyrights()),
            "description": description,
            "html_description": SpotifyMockFactory.html_description(description),
            "edition": kwargs.get("edition", random_alphanumeric_string()),
            "explicit": kwargs.get("explicit", SpotifyMockFactory._random_boolean()),
            "external_urls": SpotifyMockFactory.external_urls(entity_type, entity_id),
            "href": SpotifyMockFactory.href(entity_type, entity_id),
            "id": entity_id,
            "images": kwargs.get("images", SpotifyMockFactory.images()),
            "languages": kwargs.get("languages", SpotifyMockFactory._random_string_array()),
            "media_type": kwargs.get("media_type", random_alphanumeric_string()),
            "name": kwargs.get("name", SpotifyMockFactory.name()),
            "narrators": kwargs.get("narrators", SpotifyMockFactory.some_narrators()),
            "publisher": kwargs.get("publisher", random_alphanumeric_string()),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type, entity_id),
            "total_chapters": kwargs.get("total_chapters", randint(1, 20)),
        }

    @staticmethod
    def show(**kwargs) -> dict:
        entity_type = "show"
        entity_id = kwargs.get("id", SpotifyMockFactory.spotify_id())
        description = kwargs.get("description", SpotifyMockFactory.description())

        return {
            "available_markets": kwargs.get("available_markets", SpotifyMockFactory.available_markets()),
            "copyrights": kwargs.get("copyrights", SpotifyMockFactory.some_copyrights()),
            "description": description,
            "html_description": SpotifyMockFactory.html_description(description),
            "explicit": kwargs.get("explicit", SpotifyMockFactory._random_boolean()),
            "external_urls": SpotifyMockFactory.external_urls(entity_type, entity_id),
            "href": SpotifyMockFactory.href(entity_type, entity_id),
            "id": entity_id,
            "images": kwargs.get("images", SpotifyMockFactory.images()),
            "is_externally_hosted": kwargs.get("is_externally_hosted", SpotifyMockFactory._random_boolean()),
            "languages": kwargs.get("languages", SpotifyMockFactory._random_string_array()),
            "media_type": kwargs.get("media_type", random_alphanumeric_string()),
            "name": kwargs.get("name", SpotifyMockFactory.name()),
            "publisher": kwargs.get("publisher", random_alphanumeric_string()),
            "type": entity_type,
            "uri": SpotifyMockFactory.uri(entity_type, entity_id),
            "total_episodes": kwargs.get("total_episodes", randint(1, 20)),
        }

    @staticmethod
    def some_authors(length: Optional[int] = None) -> List[dict]:
        number_of_authors = length or randint(2, 10)
        return [SpotifyMockFactory.author() for _ in range(number_of_authors)]

    @staticmethod
    def author(**kwargs) -> dict:
        return {"name": kwargs.get("name", SpotifyMockFactory.name())}

    @staticmethod
    def some_copyrights(length: Optional[int] = None) -> List[dict]:
        number_of_copyrights = length or randint(2, 10)
        return [SpotifyMockFactory.copyright() for _ in range(number_of_copyrights)]

    @staticmethod
    def copyright(**kwargs) -> dict:
        return {
            "text": kwargs.get("text", random_alphanumeric_string()),
            "type": kwargs.get("type", choice(["C", "P"])),
        }

    @staticmethod
    def some_narrators(length: Optional[int] = None) -> List[dict]:
        number_of_authors = length or randint(2, 10)
        return [SpotifyMockFactory.author() for _ in range(number_of_authors)]

    @staticmethod
    def narrator(**kwargs) -> dict:
        return {"name": kwargs.get("name", SpotifyMockFactory.name())}

    @staticmethod
    def _optional_random_alphanumeric_string() -> Optional[str]:
        return SpotifyMockFactory._an_optional(random_alphanumeric_string)

    @staticmethod
    def _an_optional(value_generator: Callable[[], Any]) -> Optional[Any]:
        if SpotifyMockFactory._random_boolean():
            return value_generator()

    @staticmethod
    def _random_multi_enum_values(enum_: Type[EnumType]) -> List[EnumType]:
        return [v for v in enum_ if SpotifyMockFactory._random_boolean()]



from typing import List, Optional

from pytest_httpserver import RequestHandler

from spotipyio.logic.consts.typing_consts import Json
from spotipyio.testing.infra import BaseTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class CurrentProfileTestComponent(BaseTestComponent):
    def expect(self) -> List[RequestHandler]:
        return [self._create_request_handler()]

    def expect_success(self, response_json: Optional[Json] = None) -> None:
        handler = self._create_request_handler()
        handler.respond_with_json(response_json=response_json or SpotifyMockFactory.user_profile())

    def expect_failure(self, status: Optional[int] = None, response_json: Optional[Json] = None) -> None:
        status, response_json = self._create_invalid_response(status, response_json)
        handler = self._create_request_handler()

        handler.respond_with_json(status=status, response_json=response_json)

    def _create_request_handler(self) -> RequestHandler:
        return self._expect_get_request(route="/me")



from spotipyio.logic.collectors.chunks_collectors import (
    AlbumsCollector,
    ArtistsCollector,
    AudioFeaturesCollector,
    ChaptersCollector,
    EpisodesCollector,
    TracksCollector,
)
from spotipyio.logic.collectors.pagination_collectors import PlaylistsCollector, UserPlaylistsCollector
from spotipyio.logic.collectors.search_collectors import SearchCollector
from spotipyio.logic.collectors.singles_collectors import ArtistsTopTracksCollector
from spotipyio.logic.collectors.top_items_collectors import TopItemsCollector
from spotipyio.logic.collectors.current_profile_collector import CurrentProfileCollector


__all__ = [
    # Chunks
    "AlbumsCollector",
    "ArtistsCollector",
    "AudioFeaturesCollector",
    "ChaptersCollector",
    "EpisodesCollector",
    "TracksCollector",
    # Pagination
    "PlaylistsCollector",
    "UserPlaylistsCollector",
    # Other
    "SearchCollector",
    "ArtistsTopTracksCollector",
    "TopItemsCollector",
    "CurrentProfileCollector",
]



from typing import List, Optional

from pytest_httpserver import RequestHandler

from spotipyio.logic.consts.spotify_consts import PLAYLISTS, USERS
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.testing.infra import BaseTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class UserPlaylistsTestComponent(BaseTestComponent):
    def expect(self, user_id: str) -> List[RequestHandler]:
        return [self._create_request_handler(user_id)]

    def expect_success(self, user_id: str, response_json: Optional[Json] = None) -> None:
        request_handler = self._create_request_handler(user_id)
        response = response_json or SpotifyMockFactory.paged_playlists()

        request_handler.respond_with_json(response)

    def expect_failure(self, user_id: str, status: Optional[int] = None, response_json: Optional[Json] = None) -> None:
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)
        request_handler = self._create_request_handler(user_id)

        request_handler.respond_with_json(status=status, response_json=response_json)

    def _create_request_handler(self, user_id: str) -> RequestHandler:
        return self._expect_get_request(route=f"/{USERS}/{user_id}/{PLAYLISTS}", params={"offset": "0", "limit": "50"})



from typing import List, Optional

from pytest_httpserver import RequestHandler
from werkzeug import Response

from spotipyio.logic.consts.spotify_consts import PLAYLISTS, IMAGES
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.testing.infra import BaseTestComponent
from spotipyio.logic.utils import encode_image_to_base64


class PlaylistsCoverUpdaterTestComponent(BaseTestComponent):
    def expect(self, playlist_id: str, image: bytes) -> List[RequestHandler]:
        return [self._create_request_handler(playlist_id, image)]

    def expect_success(self, playlist_id: str, image: bytes) -> None:
        request_handler = self._create_request_handler(playlist_id, image)
        response = Response(status=202)

        request_handler.respond_with_response(response)

    def expect_failure(
        self, playlist_id: str, image: bytes, status: Optional[int] = None, response_json: Optional[Json] = None
    ) -> None:
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)
        request_handler = self._create_request_handler(playlist_id, image)

        request_handler.respond_with_json(status=status, response_json=response_json)

    def _create_request_handler(self, playlist_id: str, image: bytes) -> RequestHandler:
        data = encode_image_to_base64(image)
        return self._expect_put_request(route=f"/{PLAYLISTS}/{playlist_id}/{IMAGES}", data=data)



from enum import Enum


class SpotifySearchType(Enum):
    ALBUM = "album"
    ARTIST = "artist"
    AUDIOBOOK = "audiobook"
    EPISODE = "episode"
    PLAYLIST = "playlist"
    SHOW = "show"
    TRACK = "track"



from typing import List

from spotipyio.logic.contract import ISpotifyComponent
from spotipyio.auth import SpotifySession
from spotipyio.models import SearchItem
from spotipyio.logic.internal_tools import PoolExecutor


class SearchCollector(ISpotifyComponent):
    def __init__(self, base_url: str, session: SpotifySession, pool_executor: PoolExecutor = PoolExecutor()):
        super().__init__(base_url=base_url, session=session)
        self._pool_executor = pool_executor

    async def run(self, search_items: List[SearchItem]) -> List[dict]:
        return await self._pool_executor.run(iterable=search_items, func=self.run_single, expected_type=dict)

    async def run_single(self, search_item: SearchItem) -> dict:
        return await self._session.get(url=self._url, params=search_item.to_query_params())

    @property
    def _url(self) -> str:
        return f"{self._base_url}/search"



from abc import abstractmethod
from typing import Dict, Type

from multidict import CIMultiDict
from pytest_httpserver import HTTPServer

from spotipyio.testing.infra.base_test_component import BaseTestComponent


class BaseTestManager:
    def __init__(self, **named_component: Dict[str, BaseTestComponent]):
        pass

    @classmethod
    def create(cls, server: HTTPServer, headers: CIMultiDict[str]) -> "BaseTestManager":
        named_components = {}

        for name, component in cls._components().items():
            named_components[name] = component(server=server, headers=headers)

        return cls(**named_components)

    @staticmethod
    @abstractmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        raise NotImplementedError



from __future__ import annotations

from typing import Optional

from spotipyio.logic.consts.spotify_consts import SPOTIFY_API_BASE_URL
from spotipyio.auth import SpotifySession
from spotipyio.logic.managers import (
    AlbumsManager,
    ArtistsManager,
    ChaptersManager,
    CurrentUserManager,
    EpisodesManager,
    PlaylistsManager,
    SearchManager,
    TracksManager,
    UsersManager,
)


class SpotifyClient:
    def __init__(
        self,
        session: Optional[SpotifySession] = None,
        albums: Optional[AlbumsManager] = None,
        artists: Optional[ArtistsManager] = None,
        chapters: Optional[ChaptersManager] = None,
        current_user: Optional[CurrentUserManager] = None,
        episodes: Optional[EpisodesManager] = None,
        playlists: Optional[PlaylistsManager] = None,
        search: Optional[SearchManager] = None,
        tracks: Optional[TracksManager] = None,
        users: Optional[UsersManager] = None,
        base_url: str = SPOTIFY_API_BASE_URL,
    ):
        self.session = session
        self.albums = albums
        self.artists = artists
        self.chapters = chapters
        self.current_user = current_user
        self.episodes = episodes
        self.playlists = playlists
        self.search = search
        self.tracks = tracks
        self.users = users
        self._base_url = base_url

    @classmethod
    def create(cls, session: SpotifySession, base_url: str = SPOTIFY_API_BASE_URL) -> SpotifyClient:
        return SpotifyClient(
            session=session,
            artists=ArtistsManager.create(base_url, session),
            chapters=ChaptersManager.create(base_url, session),
            current_user=CurrentUserManager.create(base_url, session),
            episodes=EpisodesManager.create(base_url, session),
            playlists=PlaylistsManager.create(base_url, session),
            users=UsersManager.create(base_url, session),
            albums=AlbumsManager.create(base_url, session),
            tracks=TracksManager.create(base_url, session),
            search=SearchManager.create(base_url, session),
        )

    async def start(self) -> SpotifyClient:
        if self.session is None:
            session = SpotifySession()
            await session.start()
        else:
            session = self.session

        return SpotifyClient(
            session=session,
            artists=ArtistsManager.create(self._base_url, session),
            chapters=ChaptersManager.create(self._base_url, session),
            current_user=CurrentUserManager.create(self._base_url, session),
            episodes=EpisodesManager.create(self._base_url, session),
            playlists=PlaylistsManager.create(self._base_url, session),
            users=UsersManager.create(self._base_url, session),
            albums=AlbumsManager.create(self._base_url, session),
            tracks=TracksManager.create(self._base_url, session),
            search=SearchManager.create(self._base_url, session),
        )

    async def stop(self) -> None:
        if self.session is not None:
            await self.session.stop()

    async def __aenter__(self) -> SpotifyClient:
        return await self.start()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.stop()



from spotipyio.logic.contract import BaseSingleCollector


class ArtistsTopTracksCollector(BaseSingleCollector):
    @property
    def _route_format(self) -> str:
        return "artists/{id}/top-tracks"



from abc import ABC, abstractmethod
from typing import Optional

from spotipyio.logic.consts.typing_consts import Json
from spotipyio.auth import SpotifySession


class ISpotifyComponent(ABC):
    def __init__(self, base_url: str, session: SpotifySession):
        self._base_url = base_url
        self._session = session

    @abstractmethod
    async def run(self, *args, **kwargs) -> Optional[Json]:
        raise NotImplementedError



from spotipyio.logic.collectors.pagination_collectors.playlists_collector import PlaylistsCollector
from spotipyio.logic.collectors.pagination_collectors.user_playlists_collector import UserPlaylistsCollector

__all__ = ["PlaylistsCollector", "UserPlaylistsCollector"]



from spotipyio.auth.spotify_grant_type import SpotifyGrantType
from spotipyio.logic.consts.api_consts import GRANT_TYPE, CODE, REDIRECT_URI, JSON, REFRESH_TOKEN, CLIENT_ID


class AuthorizationPayloadBuilder:
    @staticmethod
    def build(grant_type: SpotifyGrantType, access_code: str, redirect_uri: str, client_id: str) -> dict:
        if grant_type == SpotifyGrantType.AUTHORIZATION_CODE:
            return {GRANT_TYPE: grant_type.value, CODE: access_code, REDIRECT_URI: redirect_uri, JSON: True}

        elif grant_type == SpotifyGrantType.REFRESH_TOKEN:
            return {GRANT_TYPE: grant_type.value, REFRESH_TOKEN: access_code, CLIENT_ID: client_id}

        elif grant_type == SpotifyGrantType.CLIENT_CREDENTIALS:
            return {GRANT_TYPE: grant_type.value, JSON: True}

        raise ValueError("Did not recognize grant type")



from typing import Dict, Type

from spotipyio.testing.components import UserPlaylistsTestComponent
from spotipyio.testing.infra import BaseTestManager, BaseTestComponent


class UsersTestManager(BaseTestManager):
    def __init__(self, playlists: UserPlaylistsTestComponent):
        super().__init__()
        self.playlists = playlists

    @staticmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        return {
            "playlists": UserPlaylistsTestComponent,
        }



import os
import webbrowser
from socketserver import TCPServer
from threading import Thread
from time import time
from typing import List, Optional
from urllib.parse import urlparse, urlencode

from spotipyio.logic.consts.api_consts import ACCESS_CODE_REQUEST_URL, REDIRECT_URI, CLIENT_ID, SCOPE, RESPONSE_TYPE
from spotipyio.logic.consts.env_consts import SPOTIPY_CLIENT_ID, SPOTIPY_REDIRECT_URI
from spotipyio.tools.auth.redirect_uri_request_handler import RedirectURIRequestHandler


class AccessCodeFetcher:
    def __init__(self, timeout: int = 60):
        self._timeout = timeout

    def fetch(self, scopes: List[str], client_id: Optional[str] = None, redirect_uri: Optional[str] = None) -> str:
        if redirect_uri is None:
            redirect_uri = os.environ[SPOTIPY_REDIRECT_URI]

        if client_id is None:
            client_id = os.environ[SPOTIPY_CLIENT_ID]

        try:
            return self._fetch_access_code(scopes=scopes, client_id=client_id, redirect_uri=redirect_uri)

        finally:
            RedirectURIRequestHandler.access_code = None

    def _fetch_access_code(self, scopes: List[str], client_id: str, redirect_uri: str) -> str:
        port = self._extract_redirect_uri_port(redirect_uri)
        server_thread = Thread(target=self._start_server, args=(port,), daemon=True)
        server_thread.start()
        url = self._build_authorization_url(scopes=scopes, client_id=client_id, redirect_uri=redirect_uri)
        webbrowser.open(url)

        return self._extract_access_code()

    @staticmethod
    def _extract_redirect_uri_port(redirect_uri: str) -> int:
        parsed_url = urlparse(redirect_uri)
        if parsed_url.hostname != "localhost":
            raise ValueError(
                "Invalid redirect URI! redirect_uri must be a localhost string in the following format: `http://localhost:<some-port>`"
            )

        return parsed_url.port

    @staticmethod
    def _start_server(port: int) -> None:
        with TCPServer(("", port), RedirectURIRequestHandler) as httpd:
            httpd.serve_forever()

    @staticmethod
    def _build_authorization_url(scopes: List[str], client_id: str, redirect_uri: str) -> str:
        params = {
            CLIENT_ID: client_id,
            RESPONSE_TYPE: "code",
            REDIRECT_URI: redirect_uri,
            SCOPE: " ".join(scopes),
        }
        encoded_params = urlencode(params)

        return f"{ACCESS_CODE_REQUEST_URL}?{encoded_params}"

    def _extract_access_code(self) -> str:
        start_time = time()

        while RedirectURIRequestHandler.access_code is None:
            if time() - start_time > self._timeout:
                raise TimeoutError("Access code fetching timed out!")

        return RedirectURIRequestHandler.access_code



from spotipyio.logic.consts.spotify_consts import IMAGES
from spotipyio.logic.contract import BasePlaylistsUpdater
from spotipyio.auth import SpotifySession
from spotipyio.logic.internal_tools import ImageCompressor
from spotipyio.logic.utils import encode_image_to_base64


class PlaylistCoverUpdater(BasePlaylistsUpdater):
    def __init__(self, base_url: str, session: SpotifySession, image_compressor: ImageCompressor = ImageCompressor()):
        super().__init__(base_url=base_url, session=session)
        self._image_compressor = image_compressor

    async def run(self, playlist_id: str, image: bytes, compress_if_needed: bool = True) -> None:
        url = self._build_url(playlist_id)

        if compress_if_needed:
            image = self._image_compressor.compress(image)

        if image is not None:
            data = encode_image_to_base64(image)
            await self._session.put(url=url, data=data)

    @property
    def _route(self) -> str:
        return IMAGES



from typing import Type, Dict

from spotipyio.testing.components import TopItemsTestComponent, CurrentProfileTestComponent
from spotipyio.testing.infra import BaseTestManager, BaseTestComponent


class CurrentUserTestManager(BaseTestManager):
    def __init__(self, top_items: TopItemsTestComponent, profile: CurrentProfileTestComponent):
        super().__init__()
        self.top_items = top_items
        self.profile = profile

    @staticmethod
    def _components() -> Dict[str, Type[BaseTestComponent]]:
        return {
            "top_items": TopItemsTestComponent,
            "profile": CurrentProfileTestComponent,
        }



from spotipyio.logic.managers.albums_manager import AlbumsManager
from spotipyio.logic.managers.artists_manager import ArtistsManager
from spotipyio.logic.managers.chapters_manager import ChaptersManager
from spotipyio.logic.managers.current_user_manager import CurrentUserManager
from spotipyio.logic.managers.episodes_manager import EpisodesManager
from spotipyio.logic.managers.playlists_manager import PlaylistsManager
from spotipyio.logic.managers.search_manager import SearchManager
from spotipyio.logic.managers.tracks_manager import TracksManager
from spotipyio.logic.managers.users_manager import UsersManager

__all__ = [
    "AlbumsManager",
    "ArtistsManager",
    "ChaptersManager",
    "CurrentUserManager",
    "EpisodesManager",
    "PlaylistsManager",
    "SearchManager",
    "TracksManager",
    "UsersManager",
]



from enum import Enum


class SpotifyGrantType(Enum):
    CLIENT_CREDENTIALS = "client_credentials"
    AUTHORIZATION_CODE = "authorization_code"
    REFRESH_TOKEN = "refresh_token"

    def __eq__(self, other):
        return self.value == other.value



from typing import Type, Dict

from spotipyio.logic.contract import BaseManager, ISpotifyComponent
from spotipyio.logic.collectors import EpisodesCollector


class EpisodesManager(BaseManager):
    def __init__(self, info: EpisodesCollector):
        super().__init__()
        self.info = info

    @staticmethod
    def _components() -> Dict[str, Type[ISpotifyComponent]]:
        return {
            "info": EpisodesCollector,
        }



import logging


def get_logger():
    logging.basicConfig(
        format="[%(threadName)s] %(asctime)s [%(levelname)s]  %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    return logging.getLogger("SpotipyLogger")


logger = get_logger()



from spotipyio.logic.consts.spotify_consts import ARTISTS
from spotipyio.logic.contract import BaseChunksCollector
from spotipyio.logic.internal_models import ChunkSize


class ArtistsCollector(BaseChunksCollector):
    @property
    def _route(self) -> str:
        return ARTISTS

    @property
    def _chunk_size(self) -> ChunkSize:
        return ChunkSize.ARTISTS



from base64 import b64encode
from random import randint

from spotipyio.auth import ClientCredentials, SpotifyGrantType
from spotipyio.logic.utils.general_utils import random_enum_value
from spotipyio.logic.utils.string_utils import random_alphanumeric_string
from spotipyio.models import EntityType


def to_uri(entity_id: str, entity_type: EntityType) -> str:
    return f"spotify:{entity_type.value}:{entity_id}"


def encode_bearer_token(client_id: str, client_secret: str) -> str:
    bytes_auth = bytes(f"{client_id}:{client_secret}", "ISO-8859-1")
    b64_auth = b64encode(bytes_auth)

    return b64_auth.decode("ascii")


def random_client_credentials() -> ClientCredentials:
    redirect_uri_port = randint(1000, 9999)
    return ClientCredentials(
        client_id=random_alphanumeric_string(),
        client_secret=random_alphanumeric_string(),
        redirect_uri=f"http://localhost:{redirect_uri_port}",
        grant_type=random_enum_value(SpotifyGrantType),
        access_code=random_alphanumeric_string(),
    )



from spotipyio.logic.authorization.authorization_payload_builder import AuthorizationPayloadBuilder

from spotipyio.logic.authorization.access_token_generator import AccessTokenGenerator

__all__ = ["AccessTokenGenerator", "AuthorizationPayloadBuilder"]



from enum import Enum


class ChunkSize(Enum):
    ALBUMS = 20
    ARTISTS = 50
    AUDIO_FEATURES = 100
    CHAPTERS = 50
    EPISODES = 50
    ITEMS_ADDITION = 100
    ITEMS_REMOVAL = 100
    TRACKS = 50



from abc import abstractmethod, ABC
from functools import partial
from typing import List, Optional

from spotipyio.logic.contract.spotify_component_interface import ISpotifyComponent
from spotipyio.auth import SpotifySession
from spotipyio.logic.internal_tools import PoolExecutor


class BaseSingleCollector(ISpotifyComponent, ABC):
    def __init__(self, base_url: str, session: SpotifySession, pool_executor: PoolExecutor = PoolExecutor()):
        super().__init__(base_url=base_url, session=session)
        self._pool_executor = pool_executor

    async def run(self, ids: List[str], **params) -> List[dict]:
        func = partial(self.run_single, params=params)
        return await self._pool_executor.run(iterable=ids, func=func, expected_type=dict)

    async def run_single(self, id_: str, params: Optional[dict] = None) -> dict:
        route = self._route_format.format(id=id_)
        url = f"{self._base_url}/{route}"

        return await self._session.get(url=url, params=params)

    @property
    @abstractmethod
    def _route_format(self) -> str:
        raise NotImplementedError



from __future__ import annotations

from typing import Optional
from urllib.parse import urlencode

from pytest_httpserver import HTTPServer

from spotipyio import SpotifyClient, SpotifySession
from spotipyio.auth import ClientCredentials
from spotipyio.logic.authorization import AuthorizationPayloadBuilder
from spotipyio.logic.consts.api_consts import ACCESS_TOKEN
from spotipyio.logic.utils import random_alphanumeric_string, random_client_credentials
from spotipyio.testing.managers import (
    AlbumsTestManager,
    ArtistsTestManager,
    ChaptersTestManager,
    CurrentUserTestManager,
    EpisodesTestManager,
    PlaylistsTestManager,
    SearchTestManager,
    TracksTestManager,
    UsersTestManager,
)


class SpotifyTestClient:
    def __init__(
        self,
        credentials: Optional[ClientCredentials] = None,
        api_server: Optional[HTTPServer] = None,
        authorization_server: Optional[HTTPServer] = None,
        session: Optional[SpotifySession] = None,
        albums: Optional[AlbumsTestManager] = None,
        artists: Optional[ArtistsTestManager] = None,
        chapters: Optional[ChaptersTestManager] = None,
        current_user: Optional[CurrentUserTestManager] = None,
        episodes: Optional[EpisodesTestManager] = None,
        playlists: Optional[PlaylistsTestManager] = None,
        search: Optional[SearchTestManager] = None,
        tracks: Optional[TracksTestManager] = None,
        users: Optional[UsersTestManager] = None,
    ):
        self._credentials = credentials or random_client_credentials()
        self._api_server = api_server
        self._authorization_server = authorization_server
        self._session = session
        self.albums = albums
        self.artists = artists
        self.chapters = chapters
        self.current_user = current_user
        self.episodes = episodes
        self.playlists = playlists
        self.search = search
        self.tracks = tracks
        self.users = users

    def get_base_url(self) -> str:
        return self._api_server.url_for("").rstrip("/")

    async def create_client(self) -> SpotifyClient:
        client = SpotifyClient(session=self._session, base_url=self.get_base_url())
        return await client.start()

    async def __aenter__(self) -> "SpotifyTestClient":
        self._init_api_server()
        self._init_authorization_server()
        self._expect_authorization_request()
        await self._init_session()
        self._init_managers()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._api_server is not None and self._api_server.is_running():
            self._api_server.stop()

        if self._authorization_server is not None and self._authorization_server.is_running():
            self._authorization_server.stop()

    def _init_api_server(self):
        if self._api_server is None:
            self._api_server = HTTPServer()

        if not self._api_server.is_running():
            self._api_server.start()

    def _init_authorization_server(self):
        if self._authorization_server is None:
            self._authorization_server = HTTPServer()

        if not self._authorization_server.is_running():
            self._authorization_server.start()

    def _expect_authorization_request(self) -> None:
        payload = AuthorizationPayloadBuilder.build(
            grant_type=self._credentials.grant_type,
            access_code=self._credentials.access_code,
            client_id=self._credentials.client_id,
            redirect_uri=self._credentials.redirect_uri,
        )
        request_handler = self._authorization_server.expect_request(
            uri="/", method="POST", data=bytes(urlencode(payload).encode())
        )
        authorization_server_response = {ACCESS_TOKEN: random_alphanumeric_string()}
        request_handler.respond_with_json(authorization_server_response)

    async def _init_session(self):
        if self._session is None:
            self._session = SpotifySession(
                token_request_url=self._authorization_server.url_for("").rstrip("/"),
                credentials=self._credentials,
            )

        self._session = await self._session.__aenter__()

    def _init_managers(self) -> None:
        headers = self._session.get_authorization_headers()
        self.albums = AlbumsTestManager.create(self._api_server, headers)
        self.artists = ArtistsTestManager.create(self._api_server, headers)
        self.chapters = ChaptersTestManager.create(self._api_server, headers)
        self.current_user = CurrentUserTestManager.create(self._api_server, headers)
        self.episodes = EpisodesTestManager.create(self._api_server, headers)
        self.playlists = PlaylistsTestManager.create(self._api_server, headers)
        self.search = SearchTestManager.create(self._api_server, headers)
        self.tracks = TracksTestManager.create(self._api_server, headers)
        self.users = UsersTestManager.create(self._api_server, headers)



from typing import Dict, Type

from spotipyio.logic.contract import BaseManager, ISpotifyComponent
from spotipyio.logic.collectors import PlaylistsCollector
from spotipyio.logic.creators import PlaylistsCreator
from spotipyio.logic.updaters import (
    PlaylistCoverUpdater,
    PlaylistItemsAdder,
    PlaylistItemsRemover,
    PlaylistItemsReorder,
    PlaylistItemsReplacer,
)


class PlaylistsManager(BaseManager):
    def __init__(
        self,
        add_items: PlaylistItemsAdder,
        create: PlaylistsCreator,
        info: PlaylistsCollector,
        remove_items: PlaylistItemsRemover,
        reorder_items: PlaylistItemsReorder,
        replace_items: PlaylistItemsReplacer,
        update_cover: PlaylistCoverUpdater,
    ):
        super().__init__()
        self.add_items = add_items
        self.create = create
        self.info = info
        self.remove_items = remove_items
        self.replace_items = replace_items
        self.reorder_items = reorder_items
        self.update_cover = update_cover

    @staticmethod
    def _components() -> Dict[str, Type[ISpotifyComponent]]:
        return {
            "add_items": PlaylistItemsAdder,
            "create": PlaylistsCreator,
            "info": PlaylistsCollector,
            "remove_items": PlaylistItemsRemover,
            "replace_items": PlaylistItemsReplacer,
            "reorder_items": PlaylistItemsReorder,
            "update_cover": PlaylistCoverUpdater,
        }



from typing import Type, Dict

from spotipyio.logic.collectors import TracksCollector, AudioFeaturesCollector
from spotipyio.logic.contract import BaseManager, ISpotifyComponent


class TracksManager(BaseManager):
    def __init__(self, info: TracksCollector, audio_features: AudioFeaturesCollector):
        super().__init__()
        self.info = info
        self.audio_features = audio_features

    @staticmethod
    def _components() -> Dict[str, Type[ISpotifyComponent]]:
        return {"info": TracksCollector, "audio_features": AudioFeaturesCollector}



from typing import List, Optional

from spotipyio.logic.consts.spotify_consts import URIS, TRACKS, POSITION, SNAPSHOT_ID
from spotipyio.logic.contract import BasePlaylistsUpdater
from spotipyio.auth import SpotifySession
from spotipyio.logic.internal_models import ChunkSize
from spotipyio.logic.internal_tools import DataChunksGenerator


class PlaylistItemsAdder(BasePlaylistsUpdater):
    def __init__(
        self, base_url: str, session: SpotifySession, chunks_generator: DataChunksGenerator = DataChunksGenerator()
    ):
        super().__init__(base_url=base_url, session=session)
        self._chunks_generator = chunks_generator

    async def run(self, playlist_id: str, uris: List[str], position: Optional[int] = None) -> List[str]:
        chunks = self._chunks_generator.generate_data_chunks(lst=uris, chunk_size=ChunkSize.ITEMS_ADDITION.value)
        url = self._build_url(playlist_id)
        snapshots = []

        for chunk in chunks:
            chunk_snapshot = await self._post_single_chunk(url=url, uris=chunk, position=position)
            snapshots.append(chunk_snapshot)

            if position is not None:
                position += ChunkSize.ITEMS_ADDITION.value

        return snapshots

    async def _post_single_chunk(self, url: str, uris: List[str], position: Optional[int] = None) -> str:
        payload = {URIS: uris, POSITION: position}
        response = await self._session.post(url=url, payload=payload)

        return response[SNAPSHOT_ID]

    @property
    def _route(self) -> str:
        return TRACKS



from typing import Generator, Union, Type, Any, List

from spotipyio.logic.consts.typing_consts import AF, F
from spotipyio.logic.internal_tools.pool_executor import PoolExecutor


class DataChunksGenerator:
    def __init__(self, pool_executor: PoolExecutor = PoolExecutor()):
        self._pool_executor = pool_executor

    async def execute_by_chunk_in_parallel(
        self, lst: list, func: Union[F, AF], expected_type: Type[Any], chunk_size: int
    ) -> List[Any]:
        chunks = self.generate_data_chunks(lst=lst, chunk_size=chunk_size)
        return await self._pool_executor.run(iterable=list(chunks), func=func, expected_type=expected_type)

    @staticmethod
    def generate_data_chunks(lst: list, chunk_size: int) -> Generator[list, None, None]:
        for i in range(0, len(lst), chunk_size):
            yield lst[i : i + chunk_size]



from typing import Optional

from spotipyio.logic.consts.spotify_consts import NAME, ARTISTS
from spotipyio.tools.extractors.entity_extractor_interface import IEntityExtractor


class PrimaryArtistEntityExtractor(IEntityExtractor):
    def extract(self, entity: dict) -> Optional[str]:
        items = entity.get(ARTISTS)

        if items:
            primary_artist = items[0]
            return primary_artist.get(NAME)

    @property
    def name(self) -> str:
        return "artist"



from difflib import SequenceMatcher
from random import randint, choice
from string import ascii_letters, digits


def compute_similarity_score(s1: str, s2: str) -> float:
    s1_processed = s1.lower().strip()
    s2_processed = s2.lower().strip()

    return SequenceMatcher(None, s1_processed, s2_processed).ratio()


def random_alphanumeric_string(min_length: int = 1, max_length: int = 20) -> str:
    n_chars = randint(min_length, max_length)
    characters = ascii_letters + digits

    return "".join(choice(characters) for _ in range(n_chars))



from random import randint
from typing import List
from urllib.parse import urlencode

from spotipyio.logic.consts.spotify_consts import OFFSET, LIMIT, PLAYLISTS, TRACKS
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class RandomPagedResponsesBuilder:
    def __init__(self, base_url: str, page_max_size: int):
        self._base_url = base_url
        self._page_max_size = page_max_size

    def build(self, playlist_id: str, max_pages: int) -> List[dict]:
        second_page_url = self._build_next_url(playlist_id, page_number=1)
        tracks = SpotifyMockFactory.playlist_tracks(next=second_page_url)
        playlist_response = SpotifyMockFactory.playlist(id=playlist_id, tracks=tracks, total=self._page_max_size)
        additional_pages_responses = self._build_additional_pages_responses(
            max_pages=max_pages, playlist_id=playlist_id
        )

        return [playlist_response] + additional_pages_responses

    def _build_next_url(self, playlist_id: str, page_number: int) -> str:
        params = {OFFSET: str(page_number * self._page_max_size), LIMIT: str(self._page_max_size)}
        encoded_params = urlencode(params)

        return f"{self._base_url}/{PLAYLISTS}/{playlist_id}/{TRACKS}?{encoded_params}"

    def _build_additional_pages_responses(self, max_pages: int, playlist_id: str) -> List[dict]:
        additional_pages_responses = []

        for i in range(1, max_pages):
            page_response = self._build_single_page_response(
                page_number=i + 1, max_pages=max_pages, playlist_id=playlist_id
            )
            additional_pages_responses.append(page_response)

        return additional_pages_responses

    def _build_single_page_response(self, page_number: int, max_pages: int, playlist_id: str) -> dict:
        if page_number == max_pages:
            next_url = None
            total_tracks = randint(1, self._page_max_size)
        else:
            next_url = self._build_next_url(playlist_id, page_number=page_number)
            total_tracks = self._page_max_size

        return SpotifyMockFactory.playlist_tracks(id=playlist_id, next=next_url, total=total_tracks)



from dataclasses import dataclass


@dataclass
class PlaylistCreationRequest:
    user_id: str
    name: str
    description: str
    public: bool

    def to_payload(self) -> dict:
        return {"name": self.name, "description": self.description, "public": self.public}



from typing import List, Optional

from pytest_httpserver import RequestHandler

from spotipyio.logic.consts.typing_consts import Json
from spotipyio.testing.infra.base_test_component import BaseTestComponent
from spotipyio.testing.spotify_mock_factory import SpotifyMockFactory


class ArtistsTopTracksTestComponent(BaseTestComponent):
    def expect(self, ids: List[str]) -> List[RequestHandler]:
        request_handlers = []

        for artist_id in ids:
            handler = self._create_request_handler(artist_id)
            request_handlers.append(handler)

        return request_handlers

    def expect_success(self, id_: str, response_json: Optional[Json] = None) -> None:
        response = response_json or SpotifyMockFactory.several_tracks()
        request_handler = self._create_request_handler(id_)

        request_handler.respond_with_json(response)

    def expect_failure(self, id_: str, status: Optional[int] = None, response_json: Optional[Json] = None) -> None:
        status, response_json = self._create_invalid_response(status=status, response_json=response_json)
        request_handler = self._create_request_handler(id_)

        request_handler.respond_with_json(response_json=response_json, status=status)

    def _create_request_handler(self, artist_id: str) -> RequestHandler:
        return self._expect_get_request(f"/artists/{artist_id}/top-tracks")



from spotipyio.auth.client_credentials import ClientCredentials
from spotipyio.auth.session_cache_handler_interface import ISessionCacheHandler
from spotipyio.auth.spotify_grant_type import SpotifyGrantType
from spotipyio.auth.spotify_session import SpotifySession

__all__ = [
    "ClientCredentials",
    "ISessionCacheHandler",
    "SpotifyGrantType",
    "SpotifySession",
]



from spotipyio.logic.consts.spotify_consts import TRACKS, RANGE_START, INSERT_BEFORE, SNAPSHOT_ID, RANGE_LENGTH
from spotipyio.logic.consts.typing_consts import Json
from spotipyio.logic.contract import BasePlaylistsUpdater


class PlaylistItemsReorder(BasePlaylistsUpdater):
    async def run(
        self, playlist_id: str, range_start: int, range_length: int, insert_before: int, snapshot_id: str
    ) -> Json:
        url = self._build_url(playlist_id)
        payload = {
            RANGE_START: range_start,
            INSERT_BEFORE: insert_before,
            SNAPSHOT_ID: snapshot_id,
            RANGE_LENGTH: range_length,
        }

        return await self._session.put(url=url, payload=payload)

    @property
    def _route(self) -> str:
        return TRACKS



from ssl import create_default_context
from typing import Optional

from aiohttp import ClientSession, TCPConnector, CookieJar
from certifi import where


def create_client_session(headers: Optional[dict] = None) -> ClientSession:
    ssl_context = create_default_context(cafile=where())
    return ClientSession(
        connector=TCPConnector(ssl=ssl_context), cookie_jar=CookieJar(quote_cookie=False), headers=headers
    )



from spotipyio.logic.utils.datetime_utils import get_current_timestamp
from spotipyio.logic.utils.general_utils import chain_iterable, safe_nested_get, random_enum_value, get_all_enum_values
from spotipyio.logic.utils.image_utils import read_image, encode_image_to_base64
from spotipyio.logic.utils.spotify_utils import encode_bearer_token, to_uri, random_client_credentials
from spotipyio.logic.utils.string_utils import compute_similarity_score, random_alphanumeric_string
from spotipyio.logic.utils.web_utils import create_client_session

__all__ = [
    "chain_iterable",
    "compute_similarity_score",
    "create_client_session",
    "encode_bearer_token",
    "encode_image_to_base64",
    "get_all_enum_values",
    "get_current_timestamp",
    "random_alphanumeric_string",
    "random_client_credentials",
    "random_enum_value",
    "read_image",
    "safe_nested_get",
    "to_uri",
]



import os
from dataclasses import dataclass, field
from typing import Optional

from spotipyio.auth.spotify_grant_type import SpotifyGrantType
from spotipyio.logic.consts.env_consts import SPOTIPY_CLIENT_ID, SPOTIPY_CLIENT_SECRET, SPOTIPY_REDIRECT_URI


@dataclass
class ClientCredentials:
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    redirect_uri: Optional[str] = None
    grant_type: Optional[SpotifyGrantType] = field(default_factory=lambda: SpotifyGrantType.CLIENT_CREDENTIALS)
    access_code: Optional[str] = None

    def __post_init__(self) -> None:
        self._validate_existing_credentials()
        self._validate_access_code()

    def _validate_existing_credentials(self) -> None:
        field_to_env_var_mapping = {
            "client_id": SPOTIPY_CLIENT_ID,
            "client_secret": SPOTIPY_CLIENT_SECRET,
            "redirect_uri": SPOTIPY_REDIRECT_URI,
        }

        for field_name, env_var in field_to_env_var_mapping.items():
            self._set_default_or_raise(field_name, env_var)

    def _set_default_or_raise(self, field_name: str, env_var: str) -> None:
        field_value = getattr(self, field_name)
        if field_value is not None:
            return

        default_value = os.getenv(env_var)
        if default_value is None:
            raise ValueError(
                f"Missing credential for field `{field_name}`! You must explicitly set this field or set the "
                f"`{env_var}` environment variable."
            )

        setattr(self, field_name, default_value)

    def _validate_access_code(self) -> None:
        if self.grant_type != SpotifyGrantType.CLIENT_CREDENTIALS and self.access_code is None:
            raise ValueError(
                f"Missing access code! When using the `{SpotifyGrantType.AUTHORIZATION_CODE.value}` or "
                f"`{SpotifyGrantType.REFRESH_TOKEN.value}` flows, you must pass an access code or a refresh token in "
                f"the access_code field"
            )
